"""Attribute inference attacks."""

from __future__ import annotations

import logging
import os

import matplotlib.pyplot as plt
import multiprocess as mp
import numpy as np
from fpdf import FPDF
from sklearn.base import BaseEstimator
from sklearn.preprocessing import OneHotEncoder

from sacroml.attacks import report
from sacroml.attacks.attack import Attack
from sacroml.attacks.target import Target

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

COLOR_A: str = "#86bf91"  # training set plot colour
COLOR_B: str = "steelblue"  # testing set plot colour


class AttributeAttack(Attack):
    """Attribute inference attack."""

    def __init__(
        self,
        output_dir: str = "outputs",
        write_report: bool = True,
        n_cpu: int = max(1, mp.cpu_count() - 1),
    ) -> None:
        """Construct an object to execute an attribute inference attack.

        Parameters
        ----------
        n_cpu : int
            number of CPUs used to run the attack
        output_dir : str
            name of the directory where outputs are stored
        write_report : bool
            Whether to generate a JSON and PDF report.
        """
        super().__init__(output_dir=output_dir, write_report=write_report)
        self.n_cpu = n_cpu

    def __str__(self) -> str:
        """Return the name of the attack."""
        return "Attribute inference attack"

    def attack(self, target: Target) -> dict:
        """Run attribute inference attack.

        To be used when code has access to Target class and trained target model.

        Parameters
        ----------
        target : attacks.target.Target
            target is a Target class object

        Returns
        -------
        dict
            Attack report.
        """
        if target.n_features < 1:
            logger.info("Can't run attribute inference unless features are defined.")
        else:
            logger.info("Running attribute inference attack")
            self.attack_metrics = _attribute_inference(target, self.n_cpu)
            output = self._make_report(target)
            self._write_report(output)
            return output
        return {}

    def _get_attack_metrics_instances(self) -> dict:
        """Construct the instances metric calculated, during attacks."""
        attack_metrics_experiment = {}
        attack_metrics_instances = {}
        attack_metrics_instances["instance_0"] = self.attack_metrics
        attack_metrics_experiment["attack_instance_logger"] = attack_metrics_instances
        return attack_metrics_experiment

    def _make_pdf(self, output: dict) -> FPDF:
        """Create PDF report."""
        metadata: dict = output["metadata"]
        metrics: dict = output["attack_experiment_logger"]["attack_instance_logger"][
            "instance_0"
        ]
        path: str = metadata["attack_params"]["output_dir"]
        # Create PDF
        pdf = FPDF()
        pdf.add_page()
        pdf.set_xy(0, 0)
        report.title(pdf, "Attribute Inference Attack Report")
        report.subtitle(pdf, "Introduction")
        # Add attack parameters
        report.subtitle(pdf, "Metadata")
        for key, value in metadata["attack_params"].items():
            report.line(pdf, f"{key:>30s}: {str(value):30s}", font="courier")
        # Add attack results
        report.subtitle(pdf, "Metrics")
        # Categorical
        categ_rep: list[str] = report_categorical(metrics).split("\n")
        if len(categ_rep) > 1:
            report.line(pdf, "Categorical Features:", font="courier")
            for line in categ_rep:
                report.line(pdf, line, font="courier")
        # Quantatitive
        quant_rep: list[str] = report_quantitative(metrics).split("\n")
        if len(quant_rep) > 1:
            report.line(pdf, "Quantitative Features:", font="courier")
            for line in quant_rep:
                report.line(pdf, line, font="courier")
        # Add plots
        pdf.add_page()
        report.subtitle(pdf, "Plots")
        plot_categorical_risk(metrics, path)  # Create pngs
        plot_categorical_fraction(metrics, path)
        plot_quantitative_risk(metrics, path)
        graphs = ["cat_risk.png", "cat_frac.png", "quant_risk.png"]
        for graph in graphs:
            filename = os.path.join(path, graph)
            if os.path.exists(filename):
                pdf.image(filename, x=None, y=None, w=150, h=0, type="", link="")
                os.remove(filename)
        return pdf


def _unique_max(confidences: list[float], threshold: float) -> bool:
    """Return if there is a unique maximum confidence value above threshold."""
    if len(confidences) > 0:
        max_conf = np.max(confidences)
        if max_conf < threshold:
            return False
        unique, count = np.unique(confidences, return_counts=True)
        for u, c in zip(unique, count):
            if c == 1 and u == max_conf:
                return True
    return False


def _get_inference_data(  # pylint: disable=too-many-locals
    target: Target, feature_id: int, memberset: bool
) -> tuple[np.ndarray, np.ndarray, float]:
    """Return a dataset of each sample with the attributes to test."""
    attack_feature: dict = target.features[feature_id]
    indices: list[int] = attack_feature["indices"]
    unique = np.unique(target.X_orig[:, feature_id])
    n_unique: int = len(unique)
    values = unique
    if attack_feature["encoding"] == "onehot":
        onehot_enc = OneHotEncoder()
        values = onehot_enc.fit_transform(unique.reshape(-1, 1)).toarray()
    # samples after encoding (e.g. one-hot)
    samples: np.ndarray = target.X_train
    # samples before encoding (e.g. str)
    samples_orig: np.ndarray = target.X_train_orig
    if not memberset:
        samples = target.X_test
        samples_orig = target.X_test_orig
    n_samples, x_dim = np.shape(samples)
    x_values = np.zeros((n_samples, n_unique, x_dim), dtype=np.float64)
    y_values = target.model.predict(samples)
    # for each sample to perform inference on
    # add each possible missing feature value
    for i, x in enumerate(samples):
        for j, value in enumerate(values):
            x_values[i][j] = np.copy(x)
            x_values[i][j][indices] = value
    _, counts = np.unique(samples_orig[:, feature_id], return_counts=True)
    baseline = (np.max(counts) / n_samples) * 100
    logger.debug("feature: %d x_values shape = %s", feature_id, np.shape(x_values))
    logger.debug("feature: %d y_values shape = %s", feature_id, np.shape(y_values))
    return x_values, y_values, baseline


def _infer(  # pylint: disable=too-many-locals
    target: Target,
    feature_id: int,
    threshold: float,
    memberset: bool,
) -> tuple[int, int, float, int, int]:
    """Infer attribute.

    For each possible missing value, compute the confidence scores and label
    with the target model; if the label matches the known target model label
    for the original sample, and the highest confidence score is unique, infer
    that attribute if the confidence score is greater than a threshold.
    """
    logger.debug("Attacking feature %d set %d", feature_id, int(memberset))
    correct: int = 0  # number of correct inferences made
    total: int = 0  # total number of inferences made
    x_values, y_values, baseline = _get_inference_data(target, feature_id, memberset)
    n_unique: int = len(x_values[1])
    samples = target.X_train if memberset else target.X_test
    for i, x in enumerate(x_values):  # each sample to perform inference on
        # get model confidence scores for all possible values for the sample
        confidence = target.model.predict_proba(x)
        conf = []  # confidences for each possible value with correct label
        attr = []  # features for each possible value with correct label
        # for each possible attribute value,
        # if the label matches the known target model label
        # then store the confidence score and the tested feature vector
        for j in range(n_unique):
            this_label = np.argmax(confidence[j])
            scores = confidence[j][this_label]
            if this_label == y_values[i]:
                conf.append(scores)
                attr.append(x[j])
        # is there is a unique maximum confidence score above threshold?
        if _unique_max(conf, threshold):
            total += 1
            if (attr[np.argmax(conf)] == samples[i]).all():
                correct += 1
    logger.debug("Finished attacking feature %d", feature_id)
    return correct, total, baseline, n_unique, len(samples)


def report_categorical(results: dict) -> str:
    """Return a string report of the categorical results."""
    results = results["categorical"]
    msg = ""
    for feature in results:
        name = feature["name"]
        _, _, _, n_unique, _ = feature["train"]
        msg += f"Attacking categorical feature {name} with {n_unique} unique values:\n"
        for tranche in ("train", "test"):
            correct, total, baseline, _, n_samples = feature[tranche]
            if total > 0:
                msg += (
                    f"Correctly inferred {(correct / total) * 100:.2f}% "
                    f"of {(total / n_samples) * 100:.2f}% of the {tranche} set; "
                    f"baseline: {baseline:.2f}%\n"
                )
            else:  # pragma: no cover
                msg += f"Unable to make any inferences of the {tranche} set\n"
    return msg


def report_quantitative(results: dict) -> str:
    """Return a string report of the quantitative results."""
    results = results["quantitative"]
    msg = ""
    for feature in results:
        msg += (
            f"{feature['name']}: "
            f"{feature['train']:.2f} train risk, "
            f"{feature['test']:.2f} test risk\n"
        )
    return msg


def plot_quantitative_risk(res: dict, path: str = "") -> None:
    """Generate a bar chart showing quantitative value risk scores.

    Parameters
    ----------
    res : dict
        Dictionary containing attribute inference attack results.
    path : str
        Directory to write plots.
    """
    results = res["quantitative"]
    if len(results) < 1:  # pragma: no cover
        return
    logger.debug("Plotting quantitative feature risk scores")
    x = np.arange(len(results))
    ya = []
    yb = []
    names = []
    for feature in results:
        names.append(feature["name"])
        ya.append(feature["train"] * 100)
        yb.append(feature["test"] * 100)
    fig, ax = plt.subplots(1, 1, figsize=(8, 5))
    ax.set_xticks(x)
    ax.set_xticklabels(names, rotation=90)
    ax.set_ylim([0, 100])
    ax.bar(x + 0.2, ya, 0.4, align="center", color=COLOR_A, label="train set")
    ax.bar(x - 0.2, yb, 0.4, align="center", color=COLOR_B, label="test set")
    title = "Percentage of Set at Risk for Quantitative Attributes"
    ax.set_title(f"{res['name']}\n{title}")
    ax.tick_params(axis="x", labelsize=10)
    ax.tick_params(axis="y", labelsize=10)
    ax.grid(linestyle="dotted", linewidth=1)
    ax.legend(loc="best")
    plt.margins(y=0)
    plt.tight_layout()
    filename = os.path.join(path, "quant_risk.png")
    fig.savefig(filename, pad_inches=0, bbox_inches="tight")
    logger.debug("Saved quantitative risk plot: %s", filename)


def plot_categorical_risk(  # pylint: disable=too-many-locals
    res: dict, path: str = ""
) -> None:
    """Generate a bar chart showing categorical risk scores.

    Parameters
    ----------
    res : dict
        Dictionary containing attribute inference attack results.
    path : str
        Directory to write plots.
    """
    results: list[dict] = res["categorical"]
    if len(results) < 1:  # pragma: no cover
        return
    logger.debug("Plotting categorical feature risk scores")
    x: np.ndarray = np.arange(len(results))
    ya: list[float] = []
    yb: list[float] = []
    names: list[str] = []
    for feature in results:
        names.append(feature["name"])
        correct_a, total_a, baseline_a, _, _ = feature["train"]
        correct_b, total_b, baseline_b, _, _ = feature["test"]
        a = ((correct_a / total_a) * 100) - baseline_a if total_a > 0 else 0
        b = ((correct_b / total_b) * 100) - baseline_b if total_b > 0 else 0
        ya.append(a)
        yb.append(b)
    fig, ax = plt.subplots(1, 1, figsize=(8, 5))
    ax.set_xticks(x)
    ax.set_xticklabels(names, rotation=90)
    ax.set_ylim([-100, 100])
    ax.bar(x + 0.2, ya, 0.4, align="center", color=COLOR_A, label="train set")
    ax.bar(x - 0.2, yb, 0.4, align="center", color=COLOR_B, label="test set")
    title: str = "Improvement Over Most Common Value Estimate"
    ax.set_title(f"{res['name']}\n{title}")
    ax.tick_params(axis="x", labelsize=10)
    ax.tick_params(axis="y", labelsize=10)
    ax.grid(linestyle="dotted", linewidth=1)
    ax.legend(loc="best")
    plt.margins(y=0)
    plt.tight_layout()
    filename = os.path.join(path, "cat_risk.png")
    fig.savefig(filename, pad_inches=0, bbox_inches="tight")
    logger.debug("Saved categorical risk plot: %s", filename)


def plot_categorical_fraction(  # pylint: disable=too-many-locals
    res: dict, path: str = ""
) -> None:
    """Generate a bar chart showing fraction of dataset inferred.

    Parameters
    ----------
    res : dict
        Dictionary containing attribute inference attack results.
    path : str
        Directory to write plots.
    """
    results: list[dict] = res["categorical"]
    if len(results) < 1:  # pragma: no cover
        return
    logger.debug("Plotting categorical feature tranche sizes")
    x: np.ndarray = np.arange(len(results))
    ya: list[float] = []
    yb: list[float] = []
    names: list[str] = []
    for feature in results:
        names.append(feature["name"])
        _, total_a, _, _, n_samples_a = feature["train"]
        _, total_b, _, _, n_samples_b = feature["test"]
        a = ((total_a / n_samples_a) * 100) if n_samples_a > 0 else 0
        b = ((total_b / n_samples_b) * 100) if n_samples_b > 0 else 0
        ya.append(a)
        yb.append(b)
    fig, ax = plt.subplots(1, 1, figsize=(8, 5))
    ax.set_xticks(x)
    ax.set_xticklabels(names, rotation=90)
    ax.set_ylim([0, 100])
    ax.bar(x + 0.2, ya, 0.4, align="center", color=COLOR_A, label="train set")
    ax.bar(x - 0.2, yb, 0.4, align="center", color=COLOR_B, label="test set")
    title: str = "Percentage of Set at Risk"
    ax.set_title(f"{res['name']}\n{title}")
    ax.tick_params(axis="x", labelsize=10)
    ax.tick_params(axis="y", labelsize=10)
    ax.grid(linestyle="dotted", linewidth=1)
    ax.legend(loc="best")
    plt.margins(y=0)
    plt.tight_layout()
    filename = os.path.join(path, "cat_frac.png")
    fig.savefig(filename, pad_inches=0, bbox_inches="tight")
    logger.debug("Saved categorical fraction plot: %s", filename)


def _infer_categorical(target: Target, feature_id: int, threshold: float) -> dict:
    """Return the training and test set risks of a categorical feature."""
    return {
        "name": target.features[feature_id]["name"],
        "train": _infer(target, feature_id, threshold, True),
        "test": _infer(target, feature_id, threshold, False),
    }


def _is_categorical(target: Target, feature_id: int) -> bool:
    """Return whether a feature is categorical.

    For simplicity, assumes integer datatypes are categorical.
    """
    encoding: str = target.features[feature_id]["encoding"]
    return encoding[:3] in ("str", "int") or encoding[:6] in ("onehot")


def _attack_brute_force(
    target: Target,
    features: list[int],
    n_cpu: int,
    attack_threshold: float = 0,
) -> list[dict]:
    """Perform a brute force attribute inference attack.

    Computes the target model confidence scores for every value in the list and
    makes an inference if there is a unique highest confidence score that
    exceeds attack_threshold.
    """
    logger.debug("Brute force attacking categorical features")
    args = [(target, feature_id, attack_threshold) for feature_id in features]
    with mp.Pool(processes=n_cpu) as pool:
        return pool.starmap(_infer_categorical, args)


def _get_bounds_risk_for_sample(  # pylint: disable=too-many-locals,too-many-arguments
    target_model: BaseEstimator,
    feat_id: int,
    feat_min: float,
    feat_max: float,
    sample: np.ndarray,
    c_min: float = 0,
    protection_limit: float = 0.1,
    feat_n: int = 100,
) -> bool:
    """Return whether a quantitative feature is at risk for the sample.

    Returns a bool based on conditions surrounding upper and lower bounds of
    guesses that would lead to the same model confidence.

    Parameters
    ----------
    target_model : BaseEstimator
        Trained target model.
    feat_id : int
        Index of missing feature.
    feat_min : float
        Minimum value of missing feature.
    feat_max : float
        Maximum value of missing feature.
    sample : np.ndarray
        Original known feature vector.
    c_min : float
        Defines the confidence threshold below which we say we don't care.
    protection_limit : float
        Lower [upper] bound on estimated value must not be
        above[below] lower[upper] bounds e.g. 10% of value.
    feat_n : int
        Number of attribute values to test per sample.

    Returns
    -------
    bool
        Whether the quantitative feature is at risk for the sample.
    """
    # attribute values to test - linearly sampled
    x_feat = np.linspace(feat_min, feat_max, feat_n, endpoint=True)
    # get known label
    label: int = int(target_model.predict(sample.reshape(1, -1))[0])
    # a matrix containing feature vector with linearly spaced target attribute
    x1 = np.repeat(sample.reshape(1, -1), feat_n, axis=0)
    x1[:, feat_id] = x_feat
    # get the target model confidences across the attribute range
    confidences = target_model.predict_proba(x1)
    scores = confidences[:, label]  # scores just for the model predicted label
    peak: float = np.max(scores)
    # find lowest and highest values with peak confidence
    lower_bound_index: int = 0
    while scores[lower_bound_index] < peak:
        lower_bound_index += 1
    upper_bound_index: int = feat_n - 1
    while scores[upper_bound_index] < peak:
        upper_bound_index -= 1
    # condition 1: confidence in prediction above some threshold
    # condition 2: confidence for true value == max_confidence
    # condition 3: lower bound above lower protection limit
    # condition 4: upper bound of estimate below upper protection limit
    actual_value = sample[feat_id]
    actual_probs = target_model.predict_proba(sample.reshape(1, -1))[0]
    lower_bound: float = x_feat[lower_bound_index]
    upper_bound: float = x_feat[upper_bound_index]
    return (
        peak > c_min
        and actual_probs[label] == peak
        and lower_bound >= (1 - protection_limit) * actual_value
        and upper_bound <= (1 + protection_limit) * actual_value
    )


def _get_bounds_risk_for_feature(
    target_model: BaseEstimator, feature_id: int, samples: np.ndarray
) -> float:
    """Return the average feature risk score over a set of samples."""
    feature_risk: int = 0
    n_samples: int = len(samples)
    feat_min: float = np.min(samples[:, feature_id])
    feat_max: float = np.max(samples[:, feature_id])
    for i in range(n_samples):
        sample = samples[i]
        risk = _get_bounds_risk_for_sample(
            target_model, feature_id, feat_min, feat_max, sample
        )
        if risk:  # pragma:no cover
            # can be seen working in examples
            # testing uses nursery with dummy cont. feature
            # which is not predictive
            feature_risk += 1
    return feature_risk / n_samples if n_samples > 0 else 0


def _get_bounds_risk(
    target_model: BaseEstimator,
    feature_name: str,
    feature_id: int,
    X_train: np.ndarray,
    X_test: np.ndarray,
) -> dict:
    """Return a dict containing the dataset risks of a quantitative feature."""
    return {
        "name": feature_name,
        "train": _get_bounds_risk_for_feature(target_model, feature_id, X_train),
        "test": _get_bounds_risk_for_feature(target_model, feature_id, X_test),
    }


def _get_bounds_risks(target: Target, features: list[int], n_cpu: int) -> list[dict]:
    """Compute the bounds risk for all specified features."""
    logger.debug("Computing bounds risk for all specified features")
    args = [
        (
            target.model,
            target.features[feature_id]["name"],
            feature_id,
            target.X_train,
            target.X_test,
        )
        for feature_id in features
    ]
    with mp.Pool(processes=n_cpu) as pool:
        return pool.starmap(_get_bounds_risk, args)


def _attribute_inference(target: Target, n_cpu: int) -> dict:
    """Execute attribute inference attacks on a target given a trained model."""
    # brute force attack categorical attributes using dataset unique values
    logger.info("Attacking dataset: %s", target.dataset_name)
    logger.info("Attacking categorical attributes...")
    feature_list: list[int] = []
    for feature in range(target.n_features):
        if _is_categorical(target, feature):
            feature_list.append(feature)
    results_a: list[dict] = _attack_brute_force(target, feature_list, n_cpu)
    # compute risk scores for quantitative attributes
    logger.info("Attacking quantitative attributes...")
    feature_list = []
    for feature in range(target.n_features):
        if not _is_categorical(target, feature):
            feature_list.append(feature)
    results_b: list[dict] = _get_bounds_risks(target, feature_list, n_cpu)
    # combine results into single object
    return {
        "name": target.dataset_name,
        "categorical": results_a,
        "quantitative": results_b,
    }



"""Store information about the target model and data."""

from __future__ import annotations

import logging
import os
import pickle

import numpy as np
import sklearn
import yaml

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class Target:  # pylint: disable=too-many-instance-attributes
    """Store information about the target model and data."""

    def __init__(  # pylint: disable=too-many-arguments, too-many-locals
        self,
        model: sklearn.base.BaseEstimator | None = None,
        dataset_name: str = "",
        features: dict | None = None,
        X_train: np.ndarray | None = None,
        y_train: np.ndarray | None = None,
        X_test: np.ndarray | None = None,
        y_test: np.ndarray | None = None,
        X_orig: np.ndarray | None = None,
        y_orig: np.ndarray | None = None,
        X_train_orig: np.ndarray | None = None,
        y_train_orig: np.ndarray | None = None,
        X_test_orig: np.ndarray | None = None,
        y_test_orig: np.ndarray | None = None,
        proba_train: np.ndarray | None = None,
        proba_test: np.ndarray | None = None,
    ) -> None:
        """Store information about a target model and associated data.

        Parameters
        ----------
        model : sklearn.base.BaseEstimator | None, optional
            Trained target model. Any class that implements the
            sklearn.base.BaseEstimator interface (i.e. has fit, predict and
            predict_proba methods)
        dataset_name : str
            The name of the dataset.
        features : dict
            Dictionary describing the dataset features.
        X_train : np.ndarray | None
            The (processed) training inputs.
        y_train : np.ndarray | None
            The (processed) training outputs.
        X_test : np.ndarray | None
            The (processed) testing inputs.
        y_test : np.ndarray | None
            The (processed) testing outputs.
        X_orig : np.ndarray | None
            The original (unprocessed) dataset inputs.
        y_orig : np.ndarray | None
            The original (unprocessed) dataset outputs.
        X_train_orig : np.ndarray | None
            The original (unprocessed) training inputs.
        y_train_orig : np.ndarray | None
            The original (unprocessed) training outputs.
        X_test_orig : np.ndarray | None
            The original (unprocessed) testing inputs.
        y_test_orig : np.ndarray | None
            The original (unprocessed) testing outputs.
        proba_train : np.ndarray | None
            The model predicted training probabilities.
        proba_test : np.ndarray | None
            The model predicted testing probabilities.
        """
        # Model - details
        self.model: sklearn.base.BaseEstimator | None = model
        self.model_name: str = "unknown"
        self.model_params: dict = {}
        if self.model is not None:
            self.model_name = type(self.model).__name__
            self.model_params = self.model.get_params()
        # Model - predicted probabilities
        self.proba_train: np.ndarray | None = proba_train
        self.proba_test: np.ndarray | None = proba_test
        #  Dataset - details
        self.dataset_name: str = dataset_name
        #  Dataset - processed
        self.X_train: np.ndarray | None = X_train
        self.y_train: np.ndarray | None = y_train
        self.X_test: np.ndarray | None = X_test
        self.y_test: np.ndarray | None = y_test
        self.n_samples: int = 0
        if X_train is not None and X_test is not None:
            self.n_samples = len(X_train) + len(X_test)
        #  Dataset - unprocessed
        self.X_orig: np.ndarray | None = X_orig
        self.y_orig: np.ndarray | None = y_orig
        self.X_train_orig: np.ndarray | None = X_train_orig
        self.y_train_orig: np.ndarray | None = y_train_orig
        self.X_test_orig: np.ndarray | None = X_test_orig
        self.y_test_orig: np.ndarray | None = y_test_orig
        self.n_samples_orig: int = 0
        if X_train_orig is not None and X_test_orig is not None:
            self.n_samples_orig = len(X_train_orig) + len(X_test_orig)
        self.features: dict = features if features is not None else {}
        self.n_features: int = len(self.features)
        #  Safemodel report
        self.safemodel: list = []

    def add_processed_data(
        self,
        X_train: np.ndarray,
        y_train: np.ndarray,
        X_test: np.ndarray,
        y_test: np.ndarray,
    ) -> None:
        """Add a processed and split dataset."""
        self.X_train = X_train
        self.y_train = np.array(y_train, int)
        self.X_test = X_test
        self.y_test = np.array(y_test, int)
        self.n_samples = len(X_train) + len(X_test)

    def add_feature(self, name: str, indices: list[int], encoding: str) -> None:
        """Add a feature description to the data dictionary."""
        index: int = len(self.features)
        self.features[index] = {
            "name": name,
            "indices": indices,
            "encoding": encoding,
        }
        self.n_features = len(self.features)

    def add_raw_data(  # pylint: disable=too-many-arguments
        self,
        X_orig: np.ndarray,
        y_orig: np.ndarray,
        X_train_orig: np.ndarray,
        y_train_orig: np.ndarray,
        X_test_orig: np.ndarray,
        y_test_orig: np.ndarray,
    ) -> None:
        """Add original unprocessed dataset."""
        self.X_orig = X_orig
        self.y_orig = y_orig
        self.X_train_orig = X_train_orig
        self.y_train_orig = y_train_orig
        self.X_test_orig = X_test_orig
        self.y_test_orig = y_test_orig
        self.n_samples_orig = len(X_orig)

    def _save_model(self, path: str, ext: str, target: dict) -> None:
        """Save the target model.

        Parameters
        ----------
        path : str
            Path to write the model.
        ext : str
            File extension defining the model saved format, e.g., "pkl" or "sav".
        target : dict
            Target class as a dictionary for writing yaml.
        """
        # write model
        filename: str = os.path.normpath(f"{path}/model.{ext}")
        if hasattr(self.model, "save"):
            self.model.save(filename)
        elif ext == "pkl":
            with open(filename, "wb") as fp:
                pickle.dump(self.model, fp, protocol=pickle.HIGHEST_PROTOCOL)
        else:  # pragma: no cover
            raise ValueError(f"Unsupported file format for saving a model: {ext}")
        target["model_path"] = f"model.{ext}"
        # write hyperparameters
        target["model_name"] = self.model_name
        target["model_params"] = self.model_params

    def load_model(self, model_path: str) -> None:
        """Load the target model.

        Parameters
        ----------
        model_path : str
            Path to load the model.
        """
        path = os.path.normpath(model_path)
        _, ext = os.path.splitext(path)
        if ext == ".pkl":
            with open(path, "rb") as fp:
                self.model = pickle.load(fp)
                model_type = type(self.model)
                logger.info("Loaded: %s", model_type.__name__)
        else:  # pragma: no cover
            raise ValueError(f"Unsupported file format for loading a model: {ext}")

    def _save_numpy(self, path: str, target: dict, name: str) -> None:
        """Save a numpy array variable as pickle.

        Parameters
        ----------
        path : str
            Path to save the data.
        target : dict
            Target class as a dictionary for writing yaml.
        name : str
            Name of the numpy array to save.
        """
        if getattr(self, name) is not None:
            np_path: str = os.path.normpath(f"{path}/{name}.pkl")
            target[f"{name}_path"] = f"{name}.pkl"
            with open(np_path, "wb") as fp:
                pickle.dump(getattr(self, name), fp, protocol=pickle.HIGHEST_PROTOCOL)
        else:
            target[f"{name}_path"] = ""

    def load_array(self, arr_path: str, name: str) -> None:
        """Load a data array variable from file.

        Parameters
        ----------
        arr_path : str
            Filename of a data array.
        name : str
            Name of the data array to load.
        """
        path = os.path.normpath(arr_path)
        with open(path, "rb") as fp:
            _, ext = os.path.splitext(path)
            if ext == ".pkl":
                arr = pickle.load(fp)
                setattr(self, name, arr)
                logger.info("%s shape: %s", name, arr.shape)
            else:
                raise ValueError(f"Target cannot load {ext} files.")

    def _load_array(self, arr_path: str, target: dict, name: str) -> None:
        """Load a data array variable contained in a yaml config.

        Parameters
        ----------
        arr_path : str
            Filename of a data array.
        target : dict
            Target class as a dictionary read from yaml.
        name : str
            Name of the data array to load.
        """
        key = f"{name}_path"
        if key in target and target[key] != "":
            path = f"{arr_path}/{target[key]}"
            self.load_array(path, name)

    def _save_data(self, path: str, target: dict) -> None:
        """Save the target model data.

        Parameters
        ----------
        path : str
            Path to save the data.
        target : dict
            Target class as a dictionary for writing yaml.
        """
        self._save_numpy(path, target, "X_train")
        self._save_numpy(path, target, "y_train")
        self._save_numpy(path, target, "X_test")
        self._save_numpy(path, target, "y_test")
        self._save_numpy(path, target, "X_orig")
        self._save_numpy(path, target, "y_orig")
        self._save_numpy(path, target, "X_train_orig")
        self._save_numpy(path, target, "y_train_orig")
        self._save_numpy(path, target, "X_test_orig")
        self._save_numpy(path, target, "y_test_orig")

    def _load_data(self, path: str, target: dict) -> None:
        """Load the target model data.

        Parameters
        ----------
        path : str
            Path to load the data.
        target : dict
            Target class as a dictionary read from yaml.
        """
        self._load_array(path, target, "X_train")
        self._load_array(path, target, "y_train")
        self._load_array(path, target, "X_test")
        self._load_array(path, target, "y_test")
        self._load_array(path, target, "X_orig")
        self._load_array(path, target, "y_orig")
        self._load_array(path, target, "X_train_orig")
        self._load_array(path, target, "y_train_orig")
        self._load_array(path, target, "X_test_orig")
        self._load_array(path, target, "y_test_orig")

    def _ge(self) -> float:
        """Return the model generalisation error.

        Returns
        -------
        float
            Generalisation error.
        """
        if (
            hasattr(self.model, "score")
            and self.X_train is not None
            and self.y_train is not None
            and self.X_test is not None
            and self.y_test is not None
        ):
            try:
                train = self.model.score(self.X_train, self.y_train)
                test = self.model.score(self.X_test, self.y_test)
                return test - train
            except sklearn.exceptions.NotFittedError:
                return np.NaN
        return np.NaN

    def save(self, path: str = "target", ext: str = "pkl") -> None:
        """Save the target class to persistent storage.

        Parameters
        ----------
        path : str
            Name of the output folder to save target information.
        ext : str
            File extension defining the model saved format, e.g., "pkl" or "sav".
        """
        path: str = os.path.normpath(path)
        filename: str = os.path.normpath(f"{path}/target.yaml")
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        # convert Target to dict
        target: dict = {
            "dataset_name": self.dataset_name,
            "n_samples": self.n_samples,
            "features": self.features,
            "n_features": self.n_features,
            "n_samples_orig": self.n_samples_orig,
            "generalisation_error": self._ge(),
            "safemodel": self.safemodel,
        }
        # write model and add path
        if self.model is not None:
            self._save_model(path, ext, target)
        # write data arrays and add paths
        self._save_data(path, target)
        # write yaml
        with open(filename, "w", encoding="utf-8") as fp:
            yaml.dump(target, fp, default_flow_style=False, sort_keys=False)

    def load(self, path: str = "target") -> None:
        """Load the target class from persistent storage.

        Parameters
        ----------
        path : str
            Name of the output folder containing a target yaml file.
        """
        target: dict = {}
        # load yaml
        filename: str = os.path.normpath(f"{path}/target.yaml")
        with open(filename, encoding="utf-8") as fp:
            target = yaml.safe_load(fp)
        # load parameters
        if "dataset_name" in target:
            self.dataset_name = target["dataset_name"]
            logger.info("dataset_name: %s", self.dataset_name)
        if "n_samples" in target:
            self.n_samples = target["n_samples"]
        if "features" in target:
            features: dict = target["features"]
            # convert str keys to int
            self.features = {int(key): value for key, value in features.items()}
        if "n_features" in target:
            self.n_features = target["n_features"]
            logger.info("n_features: %d", self.n_features)
        if "n_samples_orig" in target:
            self.n_samples_orig = target["n_samples_orig"]
        if "safemodel" in target:
            self.safemodel = target["safemodel"]
        # load model
        if "model_name" in target:
            self.model_name = target["model_name"]
        if "model_params" in target:
            self.model_params = target["model_params"]
        if "model_path" in target:
            model_path = os.path.normpath(f"{path}/{target['model_path']}")
            self.load_model(model_path)
        # load data
        self._load_data(path, target)

    def add_safemodel_results(self, data: list) -> None:
        """Add the results of safemodel disclosure checking.

        Parameters
        ----------
        data : list
            The results of safemodel disclosure checking.
        """
        self.safemodel = data

    def __str__(self) -> str:
        """Return the name of the dataset used."""
        return self.dataset_name



"""Supported classifiers."""

from .dp_svc import DPSVC
from .safedecisiontreeclassifier import SafeDecisionTreeClassifier
from .safekeras import SafeKerasModel
from .saferandomforestclassifier import SafeRandomForestClassifier
from .safesvc import SafeSVC
from .safetf import SafeTFModel

__all__ = [
    "DPSVC",
    "SafeDecisionTreeClassifier",
    "SafeKerasModel",
    "SafeRandomForestClassifier",
    "SafeSVC",
    "SafeTFModel",
]



"""Calculate metrics."""

from __future__ import annotations

from collections.abc import Iterable

import numpy as np
from scipy import interpolate
from scipy.stats import norm
from sklearn.metrics import confusion_matrix, roc_auc_score, roc_curve

VAR_THRESH = 1e-2


def _div(x: float, y: float, default: float) -> float:
    """Solve the problem of division by 0 and round up.

    If y is non-zero, perform x/y and round to 8dp. If it is zero, return the
    default.

    Parameters
    ----------
    x : float
        numerator
    y : float
        denominator
    default : float
        return value if y == 0

    Returns
    -------
    division : float
        x / y, or default if y == 0
    """
    return round(float(x / y), 8) if y != 0 else float(default)


def _tpr_at_fpr(
    y_true: Iterable[float],
    y_score: Iterable[float],
    fpr: float = 0.001,
    fpr_perc: bool = False,
) -> float:
    """Compute the TPR at a fixed FPR.

    In particular, returns the TPR value corresponding to a particular FPR.
    Uses interpolation to fill in gaps.

    Parameters
    ----------
    y_true : Iterable[float]
        actual class labels
    y_score : Iterable[float]
        predicted score
    fpr : float
        false positive rate at which to compute true positive rate
    fpr_perc : bool
        if the fpr is defined as a percentage

    Returns
    -------
    tpr : float
        true positive rate at fpr
    """
    if fpr_perc:
        fpr /= 100.0

    fpr_vals, tpr_vals, thresh_vals = roc_curve(y_true, y_score)
    thresh_from_fpr = interpolate.interp1d(fpr_vals, thresh_vals)
    tpr_from_thresh = interpolate.interp1d(thresh_vals, tpr_vals)

    thresh = thresh_from_fpr(fpr)
    return tpr_from_thresh(thresh)


def _expected_auc_var(auc: float, num_pos: int, num_neg: int) -> float:
    """Compute variance of AUC under assumption of uniform probabilities.

    Uses the expression given as eqn (2) in  https://cs.nyu.edu/~mohri/pub/area.pdf.

    Parameters
    ----------
    auc : float
        auc value at which to compute the variance
    num_pos : int
        number of positive examples
    num_neg : int
        number of negative examples

    Returns
    -------
    var : float
        null variance of AUC
    """
    p_xxy = p_xyy = 1 / 3
    return (
        auc * (1 - auc)
        + (num_pos - 1) * (p_xxy - auc**2)
        + (num_neg - 1) * (p_xyy - auc**2)
    ) / (num_pos * num_neg)


def min_max_disc(
    y_true: np.ndarray, pred_probs: np.ndarray, x_prop: float = 0.1, log_p: bool = True
) -> tuple[float, float, float, float]:
    """Return non-average-case methods for MIA attacks.

    Considers actual frequency of membership amongst samples with highest- and
    lowest- assessed probability of membership. If an MIA method confidently
    asserts that 5% of samples are members and 5% of samples are not, but
    cannot tell for the remaining 90% of samples, then these metrics will flag
    this behaviour, but AUC/advantage may not. Since the difference may be
    noisy, a p-value against a null of independence of true membership and
    assessed membership probability (that is, membership probabilities are
    essentially random) is also used as a metric (using a usual Gaussian
    approximation to binomial). If the p-value is low and the frequency
    difference is high (>0.5) then the MIA attack is successful for some
    samples.

    Parameters
    ----------
    y : np.ndarray
        true labels
    yp : np.ndarray
        probabilities of labels, or monotonic transformation of probabilities
    xprop : float
        proportion of samples with highest- and lowest- probability of membership to be
        considered
    logp : bool
        convert p-values to log(p).

    Returns
    -------
    maxd : float
        frequency of y=1 amongst proportion xprop of individuals with highest assessed
        membership probability
    mind : float
        frequency of y=1 amongst proportion xprop of individuals with lowest assessed
        membership probability
    mmd : float
        difference between maxd and mind
    pval : float
        p-value or log-p value corresponding to mmd against null hypothesis that random
        variables corresponding to y and yp are independent.

    Examples
    --------
    >>> y = np.random.choice(2, 100)
    >>> yp = np.random.rand(100)
    >>> maxd, mind, mmd, pval = min_max_desc(y, yp, xprop=0.2, logp=True)
    """
    n_examples = int(np.ceil(len(y_true) * x_prop))
    pos_frequency = np.mean(y_true)  # average frequency
    y_order = np.argsort(pred_probs)  # ordering permutation

    # Frequencies
    # y values corresponding to lowest k values of yp
    y_lowest_n = y_true[y_order[:n_examples]]
    # y values corresponding to highest k values of yp
    y_highest_n = y_true[y_order[-(n_examples):]]
    maxd = np.mean(y_highest_n)
    mind = np.mean(y_lowest_n)
    mmd = maxd - mind

    # P-value
    # mmd is asymptotically distributed as N(0,sdm^2) under null.
    sdm = np.sqrt(2 * pos_frequency * (1 - pos_frequency) / n_examples)
    pval = 1 - norm.cdf(mmd, loc=0, scale=sdm)  # normal CDF
    if log_p:
        pval = -115.13 if pval < 1e-50 else np.log(pval)
    # Return
    return maxd, mind, mmd, pval


def auc_p_val(auc: float, n_pos: int, n_neg: int) -> tuple[float, float]:
    """Compute the p-value for a given AUC.

    Parameters
    ----------
    auc : float
        Observed AUC value
    n_pos : int
        Number of positive examples
    n_neg : int
        Number of negative examples

    Returns
    -------
    auc_p : float
        p-value of observing an AUC > auc by chance
    auc_std : float
        standard deviation of the NULL AUC diustribution (mean = 0.5)
    """
    auc_std = np.sqrt(_expected_auc_var(0.5, n_pos, n_neg))
    auc_p = 1 - norm.cdf(auc, loc=0.5, scale=auc_std)
    return auc_p, auc_std


def _permute_rows(X_test: np.ndarray, y_test: np.ndarray) -> None:
    """Permute rows.

    Parameters
    ----------
    X_test : np.ndarray
        Array of features to be permuted.
    y_test : np.ndarray
        Array of labels to be permuted.
    """
    N, _ = X_test.shape
    order = np.random.RandomState(seed=10).permutation(N)
    X_test = X_test[order, :]
    y_test = y_test[order]


def get_metrics(  # pylint: disable=too-many-locals
    y_pred_proba: np.ndarray, y_test: np.ndarray, permute_rows: bool = True
) -> dict:
    """Calculate metrics, including attacker advantage for MIA binary.

    Implemented as Definition 4 on https://arxiv.org/pdf/1709.01604.pdf
    which is also implemented in tensorFlow-privacy
    https://github.com/tensorflow/privacy.

    Parameters
    ----------
    y_test : np.ndarray
        Test data labels.
    y_pred_proba : np.ndarray of shape [x,2] and type float
        Predicted probabilities.
    permute_rows : bool, default True
        Whether to permute arrays, see: https://github.com/AI-SDC/SACRO-ML/issues/106

    Returns
    -------
    metrics : dict
        dictionary of metric values

    Notes
    -----
    Includes the following metrics:

    * True positive rate or recall (TPR).
    * False positive rate (FPR), proportion of negative examples incorrectly \
    classified as positives.
    * False alarm rate (FAR), proportion of objects classified as positives \
    that are incorrect, also known as false discovery rate.
    * True neagative rate (TNR).
    * Positive predictive value or precision (PPV).
    * Negative predictive value (NPV).
    * False neagative rate (FNR).
    * Accuracy (ACC).
    * F1 Score - harmonic mean of precision and recall.
    * Advantage.
    """
    if len(y_pred_proba.shape) != 2:
        raise ValueError("y_pred must be an array of floats of shape [x,2]")
    if y_pred_proba.shape[1] != 2:
        raise ValueError("Metrics for multiclass classification are unsupported")

    if permute_rows:
        _permute_rows(y_pred_proba, y_test)

    y_pred = np.argmax(y_pred_proba, axis=1)
    y_pred_proba = y_pred_proba[:, 1]

    tn, fp, fn, tp = confusion_matrix(y_test, y_pred).ravel()

    metrics = {}
    # true positive rate or recall
    metrics["TPR"] = round(float(tp / (tp + fn)), 8)
    # false positive rate, proportion of negative examples incorrectly classified as positives
    metrics["FPR"] = round(float(fp / (fp + tn)), 8)
    # False alarm rate, proportion of things classified as positives that are incorrect,
    # also known as false discovery rate
    metrics["FAR"] = _div(fp, (fp + tp), 0)
    # true negative rate or specificity
    metrics["TNR"] = round(float(tn / (tn + fp)), 8)
    # precision or positive predictive value
    metrics["PPV"] = _div(tp, (tp + fp), 0)
    # negative predictive value
    metrics["NPV"] = _div(tn, (tn + fn), 0)
    # false negative rate
    metrics["FNR"] = round(float(fn / (tp + fn)), 8)
    # overall accuracy
    metrics["ACC"] = round(float((tp + tn) / (tp + fp + fn + tn)), 8)
    # harmonic mean of precision and sensitivity
    metrics["F1score"] = _div(
        2 * metrics["PPV"] * metrics["TPR"], metrics["PPV"] + metrics["TPR"], 0
    )
    # Advantage
    metrics["Advantage"] = float(abs(metrics["TPR"] - metrics["FPR"]))

    # calculate AUC of model
    metrics["AUC"] = round(roc_auc_score(y_test, y_pred_proba), 8)

    # Calculate AUC p-val
    metrics["P_HIGHER_AUC"], _ = auc_p_val(
        metrics["AUC"], y_test.sum(), len(y_test) - y_test.sum()
    )

    fmax, fmin, fdif, pdif = min_max_disc(y_test, y_pred_proba)
    metrics["FMAX01"] = fmax
    metrics["FMIN01"] = fmin
    metrics["FDIF01"] = fdif
    metrics["PDIF01"] = -pdif  # use -log(p) so answer is positive

    fmax, fmin, fdif, pdif = min_max_disc(y_test, y_pred_proba, x_prop=0.2)
    metrics["FMAX02"] = fmax
    metrics["FMIN02"] = fmin
    metrics["FDIF02"] = fdif
    metrics["PDIF02"] = -pdif  # use -log(p) so answer is positive

    fmax, fmin, fdif, pdif = min_max_disc(y_test, y_pred_proba, x_prop=0.01)
    metrics["FMAX001"] = fmax
    metrics["FMIN001"] = fmin
    metrics["FDIF001"] = fdif
    metrics["PDIF001"] = -pdif  # use -log(p) so answer is positive

    # Add some things useful for debugging / filtering
    metrics["pred_prob_var"] = y_pred_proba.var()

    # TPR at various FPR
    fpr_vals = [0.5, 0.2, 0.1, 0.01, 0.001, 0.00001]
    for fpr in fpr_vals:
        tpr = _tpr_at_fpr(y_test, y_pred_proba, fpr=fpr)
        name = f"TPR@{fpr}"
        metrics[name] = tpr

    fpr, tpr, roc_thresh = roc_curve(y_test, y_pred_proba)
    metrics["fpr"] = fpr
    metrics["tpr"] = tpr
    metrics["roc_thresh"] = roc_thresh

    metrics["n_pos_test_examples"] = y_test.sum()
    metrics["n_neg_test_examples"] = (1 - y_test).sum()

    return metrics



"""Privacy protected Keras model."""

import os
import warnings
from typing import Any

import numpy as np
import tensorflow as tf
import tensorflow_privacy as tfp
from dictdiffer import diff
from tensorflow.keras import Model as KerasModel  # pylint: disable = import-error
from tensorflow_privacy import compute_dp_sgd_privacy

from sacroml.safemodel.reporting import get_reporting_string
from sacroml.safemodel.safemodel import SafeModel

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=FutureWarning)

# this is the current class that dpvarians of optimizers come from
# may change in later versions of tensorflow_privacy
DP_CLASS_STRING = (
    "tensorflow_privacy.privacy.optimizers.dp_optimizer_keras."
    "make_keras_optimizer_class.<locals>.DPOptimizerClass"
)

DP_CLASS_STRING2 = (
    "tensorflow_privacy.privacy.optimizers.dp_optimizer_keras."
    "make_keras_generic_optimizer_class.<locals>.DPOptimizerClass"
)


def same_configs(m1: Any, m2: Any) -> tuple[bool, str]:
    """Check if two models have the same architecture."""
    num_layers = len(m1.layers)
    if len(m2.layers) != num_layers:
        errstr = get_reporting_string(name="different_layer_count")
        return False, errstr
    for layer in range(num_layers):
        m1_layer_config = m1.layers[layer].get_config()
        _ = m1_layer_config.pop("name")
        m2_layer_config = m2.layers[layer].get_config()
        _ = m2_layer_config.pop("name")
        match = list(diff(m1_layer_config, m2_layer_config, expand=True))
        num_diffs = len(match)
        if num_diffs > 0:
            msg = get_reporting_string(
                name="layer_configs_differ", layer=layer, length=num_diffs
            )
            for i in range(num_diffs):
                if match[i][0] == "change":
                    msg += get_reporting_string(
                        name="param_changed_from_to",
                        key=match[i][1],
                        val=match[i][2][0],
                        cur_val=match[i][2][1],
                    )
                else:  # should not be reachable as dense objects cannot be modified
                    msg += f"{match[i]}"  # pragma: no cover
            return False, msg

    return True, get_reporting_string(name="same_ann_config")


def same_weights(m1: Any, m2: Any) -> tuple[bool, str]:
    """Check if two nets with same architecture have the same weights."""
    num_layers = len(m1.layers)
    if num_layers != len(m2.layers):
        return False, "different numbers of layers"
    # layer 0 is input layer determined by data size
    for layer in range(1, num_layers):
        m1layer = m1.layers[layer].get_weights()
        m2layer = m2.layers[layer].get_weights()
        if len(m1layer[0][0]) != len(m2layer[0][0]):
            return False, f"layer {layer} not the same size."
        for dim in range(len(m1layer)):  # pylint: disable=consider-using-enumerate
            m1d = m1layer[dim]
            m2d = m2layer[dim]
            if not np.array_equal(m1d, m2d):  # pragma: no cover
                return False, f"dimension {dim} of layer {layer} differs"
    return True, "weights match"


def check_checkpoint_equality(v1: str, v2: str) -> tuple[bool, str]:
    """Compare two checkpoints saved with tensorflow save_model.

    On the assumption that the optimiser is not going to be saved,
    and that the model is going to be saved in frozen form
    this only checks the architecture and weights layer by layer.
    """
    msg = ""
    same = True

    try:
        model1 = tf.keras.models.load_model(v1)
    except Exception as e:  # pylint:disable=broad-except
        msg = get_reporting_string(name="error_reloading_model_v1", e=e)
        return False, msg
    try:
        model2 = tf.keras.models.load_model(v2)
    except Exception as e:  # pylint:disable=broad-except
        msg = get_reporting_string(name="error_reloading_model_v2", e=e)
        return False, msg

    same_config, config_message = same_configs(model1, model2)
    if not same_config:
        print("different config")
        msg += config_message
        same = False

    same_weight, weights_message = same_weights(model1, model2)
    if not same_weight:
        print("different weights")
        msg += weights_message
        same = False

    return same, msg


def check_dp_used(optimizer) -> tuple[bool, str]:
    """Check whether the DP optimizer was actually the one used."""
    key_needed = "_was_dp_gradients_called"
    critical_val = optimizer.__dict__.get(key_needed, "missing")

    if critical_val is True:
        reason = get_reporting_string(name="dp_optimizer_run")
        dp_used = True
    elif critical_val == "missing":
        reason = get_reporting_string(name="no_dp_gradients_key")
        dp_used = False
    elif critical_val is False:
        reason = get_reporting_string(name="changed_opt_no_fit")
        dp_used = False
    else:  # pragma: no cover
        # not currently reachable because optimizer class does
        # not support assignment
        # but leave in to future-proof
        reason = get_reporting_string(name="unrecognised_combination")
        dp_used = False

    return dp_used, reason


def check_optimizer_allowed(optimizer) -> tuple[bool, str]:
    """Check if the model's optimizer is in our white-list.

    Default setting is not allowed.
    """
    allowed = False
    opt_type = str(type(optimizer))
    reason = get_reporting_string(name="optimizer_not_allowed", optimizer=opt_type)
    if (DP_CLASS_STRING in opt_type) or (DP_CLASS_STRING2 in opt_type):
        allowed = True
        reason = get_reporting_string(name="optimizer_allowed", optimizer=opt_type)

    return allowed, reason


def check_optimizer_is_dp(optimizer) -> tuple[bool, str]:
    """Check whether optimizer is one of tensorflow's DP versions."""
    dp_used = False
    reason = "None"
    if "_was_dp_gradients_called" not in optimizer.__dict__:
        reason = get_reporting_string(name="no_dp_gradients_key")
    else:
        reason = get_reporting_string(name="found_dp_gradients_key")
        dp_used = True
    return dp_used, reason


def load_safe_keras_model(name: str = "undefined") -> tuple[bool, Any]:
    """Read model from file in appropriate format.

    Optimizer is deliberately excluded in the save.
    This is to prevent possibility of restarting training,
    which could offer possible back door into attacks.
    Thus optimizer cannot be loaded.
    """
    the_model = None
    model_load_file = name
    msg = ""
    if model_load_file == "undefined":
        msg = "Please input a name with extension for the model to load."

    elif model_load_file[-3:] == ".tf":
        # load from tf
        the_model = tf.keras.models.load_model(model_load_file)
        load = tf.keras.models.load_model(model_load_file, compile="False")
        the_model.set_weights(load.get_weights())

    else:
        suffix = model_load_file.split(".")[-1]
        msg = f"loading from a {suffix} file is currently not supported"

    if the_model is not None:
        return (True, the_model)
    return (False, msg)


class SafeKerasModel(KerasModel, SafeModel):
    """Privacy Protected Wrapper around tf.keras.Model class from tensorflow 2.8."""

    # pylint: disable=too-many-instance-attributes
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Create model and apply constraints to params."""
        the_kwargs = kwargs

        # initialise all the values that get provided as options to keras
        # and also l2 norm clipping and learning rates, batch sizes
        inputs = None
        if "inputs" in kwargs:
            inputs = the_kwargs["inputs"]
        elif len(args) == 3:  # defaults is for Model(input,outputs,names)
            inputs = args[0]
        self.outputs = None
        if "outputs" in kwargs:
            outputs = the_kwargs["outputs"]
        elif len(args) == 3:
            outputs = args[1]

        # call the keras super class first as this comes first in chain
        super().__init__(  # pylint: disable=unexpected-keyword-arg
            inputs=inputs,
            outputs=outputs,  # pylint: disable=used-before-assignment
        )

        # set values where the user has supplied them
        # if not supplied set to a value that preliminary_check
        # will over-ride with TRE-specific values from rules.json
        defaults = {
            "l2_norm_clip": 1.0,
            "noise_multiplier": 0.5,
            "min_epsilon": 10,
            "delta": 1e-5,
            "batch_size": 25,
            "num_microbatches": None,
            "learning_rate": 0.1,
            "optimizer": tfp.DPKerasSGDOptimizer,
            "num_samples": 250,
            "epochs": 20,
            "current_epsilon": 999,
        }

        for key, val in defaults.items():
            if kwargs.get(key, "missing") != "missing":
                setattr(self, key, kwargs[key])
            else:
                setattr(self, key, val)

        if self.batch_size == 0:
            msg = get_reporting_string(name="batch_size_zero")
            print(msg)
            self.batch_size = 32

        SafeModel.__init__(self)

        self.model_type: str = "KerasModel"
        # remove. this from default class
        _ = self.__dict__.pop("saved_model")
        super().preliminary_check(apply_constraints=True, verbose=True)

    def dp_epsilon_met(
        self, num_examples: int, batch_size: int = 0, epochs: int = 0
    ) -> tuple[bool, str]:
        """Check if epsilon is sufficient for Differential Privacy.

        Provides feedback to user if epsilon is not sufficient.
        """
        privacy = compute_dp_sgd_privacy(
            n=num_examples,
            batch_size=batch_size,
            noise_multiplier=self.noise_multiplier,
            epochs=epochs,
            delta=self.delta,
        )
        ok = privacy[0] < self.min_epsilon
        return ok, privacy[0]

    def check_epsilon(
        self, num_samples: int, batch_size: int, epochs: int
    ) -> tuple[bool, str]:
        """Check if the level of privacy guarantee is within recommended limits."""
        msg = ""
        ok = False
        if batch_size == 0:
            msg += get_reporting_string(name="division_by_zero")
            batch_size = 1
        (
            ok,
            self.current_epsilon,  # pylint: disable=attribute-defined-outside-init
        ) = self.dp_epsilon_met(
            num_examples=num_samples, batch_size=batch_size, epochs=epochs
        )

        key_name = "dp_requirements_met" if ok else "dp_requirements_not_met"
        get_reporting_string(
            name=key_name,
            current_epsilon=self.current_epsilon,
            num_samples=num_samples,
            batch_size=batch_size,
            epochs=epochs,
        )
        print(msg)
        return ok, msg

    def compile(self, optimizer=None, loss="categorical_crossentropy", metrics=None):
        """Compile the safe Keras model.

        Replaces the optimiser with a DP variant if needed and creates the
        necessary DP params in the opt and loss dict, then calls tf compile.
        Allow None as default value for optimizer param because we explicitly
        deal with it.
        """
        if metrics is None:
            metrics = ["accuracy"]

        replace_message = get_reporting_string(name="warn_possible_disclosure_risk")
        using_dp_sgd = get_reporting_string(name="using_dp_sgd")
        using_dp_adagrad = get_reporting_string(name="using_dp_adagrad")
        using_dp_adam = get_reporting_string(name="using_dp_adam")

        optimizer_dict = {
            None: (using_dp_sgd, tfp.DPKerasSGDOptimizer),
            tfp.DPKerasSGDOptimizer: ("", tfp.DPKerasSGDOptimizer),
            tfp.DPKerasAdagradOptimizer: ("", tfp.DPKerasAdagradOptimizer),
            tfp.DPKerasAdamOptimizer: ("", tfp.DPKerasAdamOptimizer),
            "Adagrad": (
                replace_message + using_dp_adagrad,
                tfp.DPKerasAdagradOptimizer,
            ),
            "Adam": (replace_message + using_dp_adam, tfp.DPKerasAdamOptimizer),
            "SGD": (replace_message + using_dp_sgd, tfp.DPKerasSGDOptimizer),
        }

        val = optimizer_dict.get(optimizer, "unknown")
        if val == "unknown":
            opt_msg = using_dp_sgd
            opt_used = tfp.DPKerasSGDOptimizer
        else:
            opt_msg = val[0]
            opt_used = val[1]

        self.optimizer = opt_used  # pylint: disable=attribute-defined-outside-init
        opt = opt_used(
            l2_norm_clip=self.l2_norm_clip,
            noise_multiplier=self.noise_multiplier,
            num_microbatches=self.num_microbatches,
            learning_rate=self.learning_rate,
        )

        if len(opt_msg) > 0:
            print(get_reporting_string(name="during_compilation", opt_msg=opt_msg))

        super().compile(opt, loss, metrics)

    def fit(  # pylint:disable=too-many-arguments
        self,
        X: Any,
        y: Any,
        validation_data: Any,
        epochs: int,
        batch_size: int,
        refine_epsilon: bool = False,
    ) -> Any:
        """Fit a safe Keras model.

        Overrides the tensorflow fit() method with some extra functionality:
        (i) records number of samples for checking DP epsilon values.
        (ii) does an automatic epsilon check and reports.
        (iia) if user sets refine_epsilon = true, return without fitting the model.
        (iii) then calls the tensorflow fit() function.
        (iv) finally makes a saved copy of the newly fitted model.
        """
        self.num_samples = X.shape[0]  # pylint: disable=attribute-defined-outside-init
        self.epochs = epochs  # pylint: disable=attribute-defined-outside-init
        self.batch_size = batch_size
        # make sure you are passing keywords through - but also checking batch size epochs
        ok, msg = self.check_epsilon(X.shape[0], batch_size, epochs)

        if not ok:
            print(msg)
        if refine_epsilon:
            print(
                "Not continuing with fitting model, "
                "as return epsilon was above max recommended value, "
                "and user set refine_epsilon= True"
            )
            return False, None

        returnval = super().fit(
            X,
            y,
            validation_data=validation_data,
            epochs=epochs,
            batch_size=batch_size,
        )

        # make a saved copy for later analysis
        if not os.path.exists("tfsaves"):
            os.mkdir("tfsaves")
        self.save("tfsaves/fit_model.tf")
        # pylint: disable=attribute-defined-outside-init
        self.saved_was_dpused, self.saved_reason = check_dp_used(self.optimizer)
        self.saved_epsilon = self.current_epsilon
        return returnval

    def posthoc_check(self, verbose: bool = True) -> tuple[str, bool]:
        """Check whether the model should be considered unsafe.

        For example, has been changed since fit() was last run,
        or does not meet DP policy.
        """
        disclosive = False
        msg = ""

        # have the model architecture or weights been changed?
        self.save("tfsaves/requested_model.tf")
        models_same, same_msg = check_checkpoint_equality(
            "tfsaves/fit_model.tf",
            "tfsaves/requested_model.tf",
        )
        if not models_same:
            msg += same_msg
            disclosive = True

        # was a dp-enbled optimiser provided?
        allowed, allowedmessage = check_optimizer_allowed(self.optimizer)
        if not allowed:
            msg += allowedmessage
            disclosive = True

        # was the dp-optimiser used during fit()
        dp_used, dpusedmessage = check_dp_used(self.optimizer)
        if not dp_used:
            msg += dpusedmessage
            disclosive = True

        # have values been changed since saved  immediately after fit()?
        if (
            dp_used != self.saved_was_dpused
            or dpusedmessage != self.saved_reason
            or self.saved_epsilon != self.current_epsilon
        ):
            msg += get_reporting_string(name="opt_config_changed")
            disclosive = True

        # if not what was the value of epsilon achieved
        eps_met, cur_eps = self.dp_epsilon_met(
            num_examples=self.num_samples,
            batch_size=self.batch_size,
            epochs=self.epochs,
        )
        if not eps_met:
            dpepsilonmessage = get_reporting_string(
                name="epsilon_above_normal",
                current_epsilon=cur_eps,
            )
            if verbose:
                print(
                    get_reporting_string(
                        name="recommend_further_discussion", msg=dpepsilonmessage
                    )
                )
            msg += dpepsilonmessage
            disclosive = True

        if disclosive:
            msg = get_reporting_string(name="recommend_not_release") + msg
            return msg, True

        # passed all the tests!!
        if verbose:
            msg = get_reporting_string(name="recommend_allow_release")
            msg += get_reporting_string(
                name="allow_release_eps_below_max", current_epsilon=cur_eps
            )
        return msg, False

    def save(self, name: str = "undefined") -> None:
        """Write model to file in appropriate format.

        Parameters
        ----------
        name : string
             The name of the file to save

        Notes
        -----
        Optimizer is deliberately excluded.
        To prevent possible to restart training and thus
        possible back door into attacks.
        """
        self.model_save_file = name
        while self.model_save_file == "undefined":
            print(get_reporting_string(name="input_filename_with_extension"))
            return

        thename = self.model_save_file.split(".")
        if len(thename) == 1:
            print(get_reporting_string(name="filename_must_indicate_type"))
            # "file name must indicate type as a suffix")
        else:
            suffix = self.model_save_file.split(".")[-1]

            if suffix in ("h5", "tf"):
                try:
                    tf.keras.models.save_model(
                        self,
                        self.model_save_file,
                        include_optimizer=False,
                        save_format=suffix,
                    )
                # pragma:no cover
                except Exception as er:  # pylint:disable=broad-except # pragma:no cover
                    print(  # pragma:no cover
                        get_reporting_string(
                            name="error_saving_file", suffix=suffix, er=er
                        )
                    )
            else:
                print(
                    get_reporting_string(
                        name="suffix_not_supported_for_type", model_type=self.model_type
                    )
                )



"""Generate report for TREs from JSON file."""

from __future__ import annotations

import json
import os
import pathlib
import pprint
import shutil
from datetime import date

import matplotlib.pyplot as plt
import numpy as np
import yaml


def cleanup_files_for_release(
    move_into_artefacts: list[str],
    copy_into_release: list[str],
    release_dir: str = "release_files",
    artefacts_dir: str = "training_artefacts",
) -> None:
    """Move files created during the release process into appropriate folders."""
    if not os.path.exists(release_dir):
        os.makedirs(release_dir)

    if not os.path.exists(artefacts_dir):
        os.makedirs(artefacts_dir)

    for filepath in move_into_artefacts:
        if os.path.exists(str(filepath)):
            filename = os.path.basename(filepath)
            dest = os.path.join(artefacts_dir, filename)
            shutil.move(filename, dest)

    for filepath in copy_into_release:
        if os.path.exists(str(filepath)):
            filename = os.path.basename(filepath)
            dest = os.path.join(release_dir, filename)
            shutil.copy(filepath, dest)


class GenerateJSONModule:
    """Create and append to a JSON file."""

    def __init__(self, filename: str | None = None) -> None:
        self.filename = filename

        if self.filename is None:
            self.filename = (
                "ATTACK_RESULTS" + str(date.today().strftime("%d_%m_%Y")) + ".json"
            )

        dirname = os.path.normpath(os.path.dirname(self.filename))
        os.makedirs(dirname, exist_ok=True)
        # if file doesn't exist, create it
        if not os.path.exists(self.filename):
            with open(self.filename, "w", encoding="utf-8") as f:
                f.write("")

    def add_attack_output(self, incoming_json: dict, class_name: str) -> None:
        """Add a section of JSON to the file which is already open."""
        # Read the contents of the file and then clear the file
        with open(self.filename, "r+", encoding="utf-8") as f:
            file_contents = f.read()
            file_data = json.loads(file_contents) if file_contents != "" else {}
            f.truncate(0)

        # Add the new JSON to the JSON that was in the file, and re-write
        with open(self.filename, "w", encoding="utf-8") as f:
            incoming_json = json.loads(incoming_json)

            if "log_id" in incoming_json:
                class_name = class_name + "_" + str(incoming_json["log_id"])

            file_data[class_name] = incoming_json
            json.dump(file_data, f, indent=4)

    def get_output_filename(self) -> str:
        """Return the filename of the JSON file which has been created."""
        return self.filename

    def clean_file(self) -> None:
        """Delete the file if it exists."""
        if os.path.exists(self.filename):
            os.remove(self.filename)

        with open(self.filename, "w", encoding="utf-8") as f:
            f.write("")


class AnalysisModule:
    """Wrapper module for metrics analysis modules."""

    def __init__(self) -> None:
        self.immediate_rejection = []
        self.support_rejection = []
        self.support_release = []

    def process_dict(self) -> dict:
        """Produce a risk summary output based on analysis in this module."""
        raise NotImplementedError()

    def get_recommendation(self) -> tuple:
        """Return the three recommendation buckets created by this module."""
        return self.immediate_rejection, self.support_rejection, self.support_release

    def __str__(self) -> str:
        """Return the string representation of an analysis module."""
        raise NotImplementedError()


class FinalRecommendationModule(AnalysisModule):  # pylint: disable=too-many-instance-attributes
    """Generate the first layer of a recommendation report."""

    def __init__(self, report: dict) -> None:
        super().__init__()

        self.P_VAL_THRESH = 0.05
        self.MEAN_AUC_THRESH = 0.65
        self.INSTANCE_MODEL_WEIGHTING_SCORE = 5
        self.MIN_SAMPLES_LEAF_SCORE = 5
        self.STATISTICALLY_SIGNIFICANT_SCORE = 2
        self.MEAN_AUC_SCORE = 4

        self.report = report
        self.scores = []
        self.reasons = []

    def _is_instance_based_model(self, instance_based_model_score) -> bool:
        if "model_name" in self.report:
            if self.report["model_name"] == "SVC":
                self.scores.append(instance_based_model_score)
                self.reasons.append("Model is SVM (supersedes all other tests)")
                self.immediate_rejection.append("Model is SVM")
                return True
            if self.report["model_name"] == "KNeighborsClassifier":
                self.scores.append(instance_based_model_score)
                self.reasons.append("Model is kNN (supersedes all other tests)")
                self.immediate_rejection.append("Model is kNN")
                return True
        return False

    def _tree_min_samples_leaf(self, min_samples_leaf_score: int | float) -> None:
        # Find min samples per leaf requirement
        base_path = pathlib.Path(__file__).parents[1]
        risk_appetite_path = os.path.join(base_path, "safemodel", "rules.json")
        min_samples_leaf_appetite = None

        with open(risk_appetite_path, "r+", encoding="utf-8") as f:
            file_contents = f.read()
            json_structure = json.loads(file_contents)

            rules = json_structure["DecisionTreeClassifier"]["rules"]
            for entry in rules:
                if (
                    "keyword" in entry
                    and entry["keyword"] == "min_samples_leaf"
                    and "operator" in entry
                    and entry["operator"] == "min"
                ):
                    min_samples_leaf_appetite = entry["value"]
                    break

        if (
            ("model_params" in self.report)
            and min_samples_leaf_appetite is not None
            and "min_samples_leaf" in self.report["model_params"]
        ):
            min_samples_leaf = self.report["model_params"]["min_samples_leaf"]
            if min_samples_leaf < min_samples_leaf_appetite:
                self.scores.append(min_samples_leaf_score)

                msg = "Min samples per leaf < " + str(min_samples_leaf_appetite)
                self.reasons.append(msg)
                self.support_rejection.append(msg)
            else:
                msg = "Min samples per leaf > " + str(min_samples_leaf_appetite)
                self.support_release.append(msg)

    def _statistically_significant_auc(
        self,
        p_val_thresh: float,
        mean_auc_thresh: float,
        stat_sig_score: float,
        mean_auc_score: float,
    ) -> None:
        stat_sig_auc = []
        for k in self.report:
            if (
                isinstance(self.report[k], dict)
                and "attack_experiment_logger" in self.report[k]
            ):
                for i in self.report[k]["attack_experiment_logger"][
                    "attack_instance_logger"
                ]:
                    instance = self.report[k]["attack_experiment_logger"][
                        "attack_instance_logger"
                    ][i]

                    auc_key = "P_HIGHER_AUC"
                    if auc_key in instance and instance[auc_key] < p_val_thresh:
                        stat_sig_auc.append(instance["AUC"])

                n_instances = len(
                    self.report[k]["attack_experiment_logger"]["attack_instance_logger"]
                )
                if len(stat_sig_auc) / n_instances > 0.1:
                    msg = ">10% AUC are statistically significant in experiment " + str(
                        k
                    )

                    self.scores.append(stat_sig_score)
                    self.reasons.append(msg)
                    self.support_rejection.append(msg)
                else:
                    msg = "<10% AUC are statistically significant in experiment " + str(
                        k
                    )
                    self.support_release.append(msg)

                if len(stat_sig_auc) > 0:
                    mean = np.mean(np.array(stat_sig_auc))
                    if mean > mean_auc_thresh:
                        msg = "Attack AUC > threshold of " + str(mean_auc_thresh)
                        msg = msg + " in experiment " + str(k)

                        self.scores.append(mean_auc_score)
                        self.reasons.append(msg)
                        self.support_rejection.append(msg)
                    else:
                        msg = "Attack AUC <= threshold of " + str(mean_auc_thresh)
                        msg = msg + " in experiment " + str(k)
                        self.support_release.append(msg)

    def process_dict(self) -> dict:
        """Return a dictionary summarising the metrics."""
        self._tree_min_samples_leaf(self.MIN_SAMPLES_LEAF_SCORE)
        self._statistically_significant_auc(
            self.P_VAL_THRESH,
            self.MEAN_AUC_THRESH,
            self.STATISTICALLY_SIGNIFICANT_SCORE,
            self.MEAN_AUC_SCORE,
        )

        if len(self.scores) == 0:
            summarised_score = 0
        else:
            summarised_score = int(np.sum(np.array(self.scores)).round(0))
            summarised_score = min(summarised_score, 5)

        # if model is instance based, it is automatically disclosive. Assign max score
        if self._is_instance_based_model(self.INSTANCE_MODEL_WEIGHTING_SCORE):
            summarised_score = self.INSTANCE_MODEL_WEIGHTING_SCORE

        return {}

    def __str__(self) -> str:
        """Return string representation of the final recommendation."""
        return "Final Recommendation"


class SummariseUnivariateMetricsModule(AnalysisModule):
    """Summarise a set of chosen univariate metrics from the output dictionary."""

    def __init__(self, report: dict, metrics_list=None) -> None:
        super().__init__()

        if metrics_list is None:
            metrics_list = ["AUC", "ACC", "FDIF01"]

        self.report = report
        self.metrics_list = metrics_list

    def process_dict(self) -> dict:
        """Return a dictionary summarising the metrics."""
        output_dict = {}

        for k in self.report:
            if (
                isinstance(self.report[k], dict)
                and "attack_experiment_logger" in self.report[k]
            ):
                metrics_dict = {m: [] for m in self.metrics_list}
                for _, iteration_value in self.report[k]["attack_experiment_logger"][
                    "attack_instance_logger"
                ].items():
                    for m in metrics_dict:
                        metrics_dict[m].append(iteration_value[m])
                output = {}
                for m in self.metrics_list:
                    output[m] = {
                        "min": min(metrics_dict[m]),
                        "max": max(metrics_dict[m]),
                        "mean": np.mean(metrics_dict[m]),
                        "median": np.median(metrics_dict[m]),
                    }
                output_dict[k] = output
        return output_dict

    def __str__(self) -> str:
        """Return the string representation of a univariate metrics module."""
        return "Summary of Univarite Metrics"


class SummariseAUCPvalsModule(AnalysisModule):
    """Summarise a list of AUC values."""

    def __init__(
        self, report: dict, p_thresh: float = 0.05, correction: str = "bh"
    ) -> None:
        super().__init__()

        self.report = report
        self.p_thresh = p_thresh
        self.correction = correction

    def _n_sig(self, p_val_list: list[float], correction: str = "none") -> int:
        """Compute the number of significant p-vals with different corrections."""
        if correction == "none":
            return len(np.where(np.array(p_val_list) <= self.p_thresh)[0])
        if correction == "bh":
            sorted_p = sorted(p_val_list)
            m = len(p_val_list)
            alpha = self.p_thresh
            comparators = np.arange(1, m + 1, 1) * alpha / m
            return (sorted_p <= comparators).sum()
        if correction == "bo":  # bonferroni
            return len(
                np.where(np.array(p_val_list) <= self.p_thresh / len(p_val_list))[0]
            )

        raise NotImplementedError()  # any others?

    def _get_metrics_list(self) -> list[float]:
        metrics_list = []
        for k in self.report:
            if (
                isinstance(self.report[k], dict)
                and "attack_experiment_logger" in self.report[k]
            ):
                for _, iteration_value in self.report[k]["attack_experiment_logger"][
                    "attack_instance_logger"
                ].items():
                    metrics_list.append(iteration_value["P_HIGHER_AUC"])
        return metrics_list

    def process_dict(self) -> dict:
        """Process the dict to summarise the number of significant AUC p-values."""
        p_val_list = self._get_metrics_list()
        return {
            "n_total": len(p_val_list),
            "p_thresh": self.p_thresh,
            "n_sig_uncorrected": self._n_sig(p_val_list),
            "correction": self.correction,
            "n_sig_corrected": self._n_sig(p_val_list, self.correction),
        }

    def __str__(self) -> str:
        """Return the string representation of a AUC p-values module."""
        return f"Summary of AUC p-values at p = ({self.p_thresh})"


class SummariseFDIFPvalsModule(SummariseAUCPvalsModule):
    """Summarise the number of significant FDIF p-values."""

    def get_metric_list(self, input_dict: dict) -> list[float]:
        """Get metrics_list from attack_instance_logger within JSON file."""
        metric_list = []
        for _, iteration_value in input_dict["attack_instance_logger"].items():
            metric_list.append(iteration_value["PDIF01"])
        return [np.exp(-m) for m in metric_list]

    def __str__(self) -> str:
        """Return the string representation of a FDIF p-values module."""
        return f"Summary of FDIF p-values at p = ({self.p_thresh})"


class LogLogROCModule(AnalysisModule):
    """Generate a log-log plot."""

    def __init__(
        self, report: dict, output_folder: str | None = None, include_mean: bool = True
    ) -> None:
        super().__init__()

        self.report = report
        self.output_folder = output_folder
        self.include_mean = include_mean

    def process_dict(self) -> str:
        """Create a roc plot for multiple repetitions."""
        log_plot_names = []

        for k in self.report:
            if (
                isinstance(self.report[k], dict)
                and "attack_experiment_logger" in self.report[k]
            ):
                plt.figure(figsize=(8, 8))
                plt.plot([0, 1], [0, 1], "k--")

                # Compute average ROC
                base_fpr = np.linspace(0, 1, 1000)
                metrics = self.report[k]["attack_experiment_logger"][
                    "attack_instance_logger"
                ].values()
                all_tpr = np.zeros((len(metrics), len(base_fpr)), float)

                for i, metric_set in enumerate(metrics):
                    all_tpr[i, :] = np.interp(
                        base_fpr, metric_set["fpr"], metric_set["tpr"]
                    )

                for _, metric_set in enumerate(metrics):
                    plt.plot(
                        metric_set["fpr"],
                        metric_set["tpr"],
                        color="lightsalmon",
                        linewidth=0.5,
                    )

                tpr_mu = all_tpr.mean(axis=0)
                plt.plot(base_fpr, tpr_mu, "r")

                plt.xlabel("False Positive Rate")
                plt.ylabel("True Positive Rate")
                plt.xscale("log")
                plt.yscale("log")
                plt.tight_layout()
                plt.grid()
                out_file = (
                    f"{self.report['log_id']}-{self.report['metadata']['attack']}.png"
                )
                if self.output_folder is not None:
                    out_file = os.path.join(self.output_folder, out_file)
                plt.savefig(out_file)
                log_plot_names.append(out_file)
        return "Log plot(s) saved to " + str(log_plot_names)

    def __str__(self) -> str:
        """Return the string representation of a ROC log plot module."""
        return "ROC Log Plot"


class GenerateTextReport:
    """Generate a text report from a JSON input."""

    def __init__(self) -> None:
        self.text_out = []
        self.target_yaml_filename = None
        self.attack_json_filename = None
        self.model_name_from_target = None

        self.immediate_rejection = []
        self.support_rejection = []
        self.support_release = []

    def _process_target_yaml(self) -> None:
        """Create a summary of a target model YAML file."""
        model_params_of_interest = [
            "C",
            "kernel",
            "n_neighbors",
            "hidden_layer_sizes",
            "activation",
            "max_depth",
            "min_samples_split",
            "min_samples_leaf",
            "n_estimators",
            "learning_rate",
        ]

        with open(self.target_yaml_filename, encoding="utf-8") as f:
            yaml_report = yaml.safe_load(f)

        output_string = "TARGET MODEL SUMMARY\n"

        if "model_name" in yaml_report:
            output_string = (
                output_string + "model_name: " + yaml_report["model_name"] + "\n"
            )

        if "n_samples" in yaml_report:
            output_string = output_string + "number of samples used to train: "
            output_string = output_string + str(yaml_report["n_samples"]) + "\n"

        if "model_params" in yaml_report:
            for param in model_params_of_interest:
                if param in yaml_report["model_params"]:
                    output_string = output_string + param + ": "
                    output_string = output_string + str(
                        yaml_report["model_params"][param]
                    )
                    output_string = output_string + "\n"

        if "model_path" in yaml_report:
            filepath = os.path.split(os.path.abspath(self.target_yaml_filename))[0]
            self.model_name_from_target = os.path.join(
                filepath, yaml_report["model_path"]
            )

        self.text_out.append(output_string)

    def pretty_print(self, report: dict, title: str) -> str:
        """Format JSON code to make it more readable for TREs."""
        returned_string = title + "\n"
        for key in report:
            returned_string = returned_string + key + "\n"
            returned_string = returned_string + pprint.pformat(report[key]) + "\n\n"
        return returned_string

    def process_attack_target_json(
        self, attack_filename: str, target_filename: str = None
    ) -> None:
        """Create a neat summary of an attack JSON file."""
        self.attack_json_filename = attack_filename

        with open(attack_filename, encoding="utf-8") as f:
            json_report = json.loads(f.read())

        if target_filename is not None:
            self.target_yaml_filename = target_filename

            with open(target_filename, encoding="utf-8") as f:
                target_file = yaml.safe_load(f)
                json_report = {**json_report, **target_file}

        modules = [
            FinalRecommendationModule(json_report),
        ]

        for m in modules:
            output = m.process_dict()
            returned = m.get_recommendation()

            self.immediate_rejection += returned[0]
            self.support_rejection += returned[1]
            self.support_release += returned[2]

        output_string = self.pretty_print(output, "ATTACK JSON RESULTS")

        self.text_out.append(output_string)

        bucket_text = "Immediate rejection recommended for the following reason:\n"
        if len(self.immediate_rejection) > 0:
            for reason in self.immediate_rejection:
                bucket_text += str(reason) + "\n"
        else:
            bucket_text += "None\n"

        bucket_text += "\nEvidence supporting rejection:\n"
        if len(self.support_rejection) > 0:
            for reason in self.support_rejection:
                bucket_text += str(reason) + "\n"
        else:
            bucket_text += "None\n"

        bucket_text += "\nEvidence supporting release:\n"
        if len(self.support_release) > 0:
            for reason in self.support_release:
                bucket_text += str(reason) + "\n"
        else:
            bucket_text += "None\n"

        self.text_out.append(bucket_text)

    def export_to_file(  # pylint: disable=too-many-arguments
        self,
        output_filename: str = "summary.txt",
        move_files: bool = False,
        model_filename: str | None = None,
        release_dir: str = "release_files",
        artefacts_dir: str = "training_artefacts",
    ) -> None:
        """Take the input strings collected and combine into a neat text file."""
        copy_of_text_out = self.text_out
        self.text_out = []

        if self.target_yaml_filename is not None:
            self._process_target_yaml()

        self.text_out += copy_of_text_out

        output_filename = output_filename.replace(" ", "_")

        with open(output_filename, "w", encoding="utf-8") as text_file:
            for output_string in self.text_out:
                text_file.write(output_string)
                text_file.write("\n")

        if move_files is True:
            move_into_artefacts = ["log_roc.png"]

            copy_into_release = [
                output_filename,
                self.attack_json_filename,
                self.target_yaml_filename,
            ]

            if model_filename is None:
                copy_into_release.append(self.model_name_from_target)
            else:
                copy_into_release.append(model_filename)

            cleanup_files_for_release(
                move_into_artefacts, copy_into_release, release_dir, artefacts_dir
            )



"""Privacy protected Support Vector Classifier."""

from __future__ import annotations

import copy

import numpy as np
from dictdiffer import diff

from sacroml.safemodel.safemodel import SafeModel

from .dp_svc import DPSVC


class SafeSVC(SafeModel, DPSVC):
    """Privacy protected Support Vector Classifier."""

    def __init__(
        self,
        C: float = 1.0,
        gamma: str | float = "scale",
        dhat: int = 1000,
        eps: float = 10,
        **kwargs: dict,
    ) -> None:
        """Initialise a differentially private SVC."""
        SafeModel.__init__(self)
        DPSVC.__init__(self, C=C, gamma=gamma, dhat=dhat, eps=eps, **kwargs)
        self.model_type: str = "SVC"
        self.ignore_items = [
            "model_save_file",
            "ignore_items",
            "train_features",
            "train_labels",
            "unique_labels",
            "train_labels",
            "weights",
            "noisy_weights",
        ]
        self.examine_seperately_items = ["platt_transform", "svc"]

    def fit(self, train_features: np.ndarray, train_labels: np.ndarray) -> None:
        """Fit model and store model dict."""
        super().fit(train_features, train_labels)
        self.saved_model = copy.deepcopy(self.__dict__)

    def additional_checks(
        self, curr_separate: dict, saved_separate: dict
    ) -> tuple[str, str]:
        """SVC specific checks."""
        msg = ""
        disclosive = False
        for item in self.examine_seperately_items:
            diffs_list = list(
                diff(curr_separate[item].__dict__, saved_separate[item].__dict__)
            )
            if len(diffs_list) > 0:
                disclosive = True
                if len(diffs_list) == 1:
                    msg += f"structure {item} has one difference.\n"
                else:
                    msg += f"structure {item} has several differences.\n"

        return msg, disclosive



"""Base class for an attack object."""

from __future__ import annotations

import importlib
import inspect
import logging
import os
import uuid
from datetime import datetime

from fpdf import FPDF

from sacroml.attacks import report
from sacroml.attacks.target import Target

logger = logging.getLogger(__name__)


class Attack:
    """Base class to represent an attack."""

    def __init__(self, output_dir: str = "outputs", write_report: bool = True) -> None:
        """Instantiate an attack.

        Parameters
        ----------
        output_dir : str
            name of the directory where outputs are stored
        write_report : bool
            Whether to generate a JSON and PDF report.
        """
        self.output_dir: str = output_dir
        self.write_report: bool = write_report
        self.attack_metrics: dict | list = {}
        self.metadata: dict = {}
        if not os.path.exists(self.output_dir):
            os.makedirs(self.output_dir)

    def attack(self, target: Target) -> dict:
        """Run an attack."""
        raise NotImplementedError

    def _construct_metadata(self) -> None:
        """Generate attack metadata."""
        self.metadata = {
            "attack_name": str(self),
            "attack_params": self.get_params(),
            "global_metrics": {},
        }

    def _get_attack_metrics_instances(self) -> dict:
        """Get metrics for each individual repetition of an attack."""
        raise NotImplementedError  # pragma: no cover

    def _make_pdf(self, output: dict) -> FPDF | None:
        """Create PDF report."""
        raise NotImplementedError  # pragma: no cover

    def _make_report(self, target: Target) -> dict:
        """Create attack report."""
        logger.info("Generating report")
        self._construct_metadata()
        self.metadata["target_model"] = target.model_name
        self.metadata["target_model_params"] = target.model_params
        output: dict = {
            "log_id": str(uuid.uuid4()),
            "log_time": datetime.now().strftime("%d/%m/%Y %H:%M:%S"),
            "metadata": self.metadata,
            "attack_experiment_logger": self._get_attack_metrics_instances(),
        }
        return output

    def _write_report(self, output: dict) -> None:
        """Write report as JSON and PDF."""
        dest: str = os.path.join(self.output_dir, "report")
        if self.write_report:
            logger.info("Writing report: %s.json %s.pdf", dest, dest)
            report.write_json(output, dest)
            pdf_report = self._make_pdf(output)
            if pdf_report is not None:
                report.write_pdf(dest, pdf_report)

    def __str__(self) -> str:
        """Return the string representation of an attack."""
        raise NotImplementedError

    @classmethod
    def _get_param_names(cls) -> list[str]:
        """Get parameter names."""
        init_signature = inspect.signature(cls.__init__)
        return [
            p.name
            for p in init_signature.parameters.values()
            if p.name != "self" and p.kind != p.VAR_KEYWORD
        ]

    def get_params(self) -> dict:
        """Get parameters for this attack.

        Returns
        -------
        params : dict
            Parameter names mapped to their values.
        """
        out: dict = {}
        for key in self._get_param_names():
            out[key] = getattr(self, key)
        return out


def get_class_by_name(class_path: str):
    """Return a class given its name."""
    module_path, class_name = class_path.rsplit(".", 1)
    module = importlib.import_module(module_path)
    return getattr(module, class_name)



"""Differentially private SVC."""

from __future__ import annotations

import logging

import numpy as np
from sklearn.linear_model import LogisticRegression
from sklearn.svm import SVC

local_logger = logging.getLogger(__file__)
local_logger.setLevel(logging.WARNING)

SMALL_NUMBER = 1e-16  # used to set gamma value if zero to avoid divide by zero

# pylint: disable=too-many-instance-attributes


class DPSVC:
    """Differentially private SVM.

    Implemented according to: https://arxiv.org/pdf/0911.5708.pdf.

    Essentially approximates an infinite-dimensional latent space (and
    corresponding kernel) with a finite dimensional latent space, and adds
    noise to the normal to the separating hyperplane in this latent space.

    Only currently implemented for a radial basis kernel, but could be extended.

    More specifically
    - draws a set of dhat random vectors from a probability measure induced by
      the Fourier transform of the kernel function
    - approximates the kernel with a 2*dhat dimensional latent space
    - computes the separating hyperplane in this latent space with normal w
    - then adds Laplacian noise to w and returns it along with the map to the
      latent space.

    The SKlearn SVM (see
    https://scikit-learn.org/stable/modules/svm.html#mathematical-formulation)
    minimises the function

    (1/2) ||w||_2 + C sum(zeta_i)

    where 1-zeta_i≤ y_i (w phi(x_i) + b), where phi maps x to the latent space
    and zeta_i ≥ 0.

    This is equivalent to minimising

    (1/2) ||w||_2 + C/n sum(l(y_i,f_w(x_i)))

    where l(x,y)=n*max(0,1- x.y), which is n-Lipschitz continuous in y (given x
    is in {-1,1})
    """

    def __init__(
        self,
        C: float = 1.0,
        gamma: str | float = "scale",
        dhat: int = 1000,
        eps: float = 10,
        **kwargs: dict,
    ) -> None:
        self.svc = None
        self.gamma = gamma
        self.dpsvc_gamma = None
        self.dhat = dhat
        self.eps = eps
        self.C = C
        self.lambdaval = None
        self.rho = None
        self.support = None
        self.platt_transform = LogisticRegression()
        self.b = None
        self.classes_ = [0, 1]
        self.intercept = None
        self.noisy_weights = None
        self.set_params(**kwargs)

    def phi_hat(self, input_vector: np.ndarray) -> np.ndarray:
        """Project a single feature."""
        vt1 = (self.rho * input_vector).sum(axis=1)
        vt = (self.dhat ** (-0.5)) * np.column_stack((np.cos(vt1), np.sin(vt1)))
        return vt.reshape(2 * self.dhat)

    def phi_hat_multi(self, input_features: np.ndarray) -> np.ndarray:
        """Compute feature space for a matrix of inputs."""
        n_data, _ = input_features.shape
        phi_hat = np.zeros((n_data, 2 * self.dhat), float)
        for i in range(n_data):
            phi_hat[i, :] = self.phi_hat(input_features[i, :])
        return phi_hat

    def k_hat_svm(self, x: np.ndarray, y: np.ndarray | None = None) -> np.ndarray:
        """Define the version which is sent to sklearn.svm."""
        phi_hat_x = self.phi_hat_multi(x)
        phi_hat_y = phi_hat_x if y is None else self.phi_hat_multi(y)
        return np.dot(phi_hat_x, phi_hat_y.T)

    def fit(self, train_features: np.ndarray, train_labels: np.ndarray) -> None:
        """Fit the model."""
        # Check that the data passed is np.ndarray
        if not isinstance(train_features, np.ndarray) or not isinstance(
            train_labels, np.ndarray
        ):
            raise NotImplementedError("DPSCV needs np.ndarray inputs")

        n_data, n_features = train_features.shape

        # Check the data passed in train_labels
        unique_labels = np.unique(train_labels)
        local_logger.info(unique_labels)
        for label in unique_labels:
            if label not in [0, 1]:
                raise NotImplementedError(
                    (
                        "DP SVC can only handle binary classification with",
                        "labels = 0 and 1",
                    )
                )

        if self.eps > 0:
            self.lambdaval = (2**2.5) * self.C * np.sqrt(self.dhat) / self.eps
        else:
            self.lambdaval = 0

        # Mimic sklearn skale and auto params
        if self.gamma == "scale":
            self.gamma = 1.0 / (n_features * train_features.var())
        elif self.gamma == "auto":
            self.gamma = 1.0 / n_features

        if self.gamma == 0.0:
            self.gamma = SMALL_NUMBER
            local_logger.warning(
                "gamma value passed in was zero, set to %g", SMALL_NUMBER
            )
        # alternative parameterisation
        self.dpsvc_gamma = 1.0 / np.sqrt(2.0 * self.gamma)

        local_logger.info(
            "Gamma = %f (dp parameterisation = %f)", self.gamma, self.dpsvc_gamma
        )

        # Draw dhat random vectors rho from Fourier transform of RBF
        # (which is Gaussian with SD 1/gamma)
        self.rho = np.random.normal(0, 1.0 / self.dpsvc_gamma, (self.dhat, n_features))
        local_logger.info("Sampled rho")

        # Fit support vector machine
        # Create the gram matrix to pass to SVC
        gram_matrix = self.k_hat_svm(train_features)
        local_logger.info("Fitting base SVM")
        self.svc = SVC(kernel="precomputed", C=self.C)
        self.svc.fit(gram_matrix, train_labels)

        # Get separating hyperplane and intercept
        # alpha from solved dual, multiplied by labels (-1,1)
        alpha = self.svc.dual_coef_
        xi = train_features[self.svc.support_, :]  # support vectors x_i
        weights = np.zeros(2 * self.dhat)
        for i in range(alpha.shape[1]):
            weights = weights + alpha[0, i] * self.phi_hat(xi[i, :])

        self.intercept = self.svc.intercept_

        # Add Laplacian noise
        self.noisy_weights = weights + np.random.laplace(
            0, self.lambdaval, len(weights)
        )

        # Logistic transform for predict_proba (rough): generate predictions
        # (DP) for training data
        ypredn = np.zeros(n_data)
        for i in range(n_data):
            ypredn[i] = (
                np.dot(self.phi_hat(train_features[i, :]), self.noisy_weights)
                + self.intercept
            )

        local_logger.info("Fitting Platt scaling")
        self.platt_transform.fit(ypredn.reshape(-1, 1), train_labels)

    def set_params(self, **kwargs: dict) -> None:
        """Set params."""
        for key, value in kwargs.items():
            if key == "gamma":
                self.gamma = value
            elif key == "eps":
                self.eps = value
            elif key == "dhat":
                self.dhat = value
            else:
                local_logger.warning("Unsupported parameter: %s", key)

    def _raw_outputs(self, test_features: np.ndarray) -> np.ndarray:
        """Get the raw output, used by predict and predict_proba."""
        projected_features = self.phi_hat_multi(test_features)
        return np.dot(projected_features, self.noisy_weights) + self.intercept

    def predict(self, test_features: np.ndarray) -> np.ndarray:
        """Return the predictions."""
        out = self._raw_outputs(test_features)
        return 1 * (out > 0)

    def predict_proba(self, test_features: np.ndarray) -> np.ndarray:
        """Return the predictive probabilities."""
        out = self._raw_outputs(test_features)
        return self.platt_transform.predict_proba(out.reshape(-1, 1))



"""Structural attacks.

Runs a number of 'static' structural attacks based on:
(i) the target model's properties;
(ii) the TREs risk appetite as applied to tables and standard regressions.

Tree-based model types currently supported.
"""

from __future__ import annotations

import logging

import numpy as np
from acro import ACRO
from sklearn.base import BaseEstimator
from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.tree import DecisionTreeClassifier
from xgboost.sklearn import XGBClassifier

from sacroml.attacks.attack import Attack
from sacroml.attacks.target import Target

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# pylint: disable=chained-comparison


def get_unnecessary_risk(model: BaseEstimator) -> bool:
    """Check whether model hyperparameters are in the top 20% most risky.

    This check is designed to assess whether a model is likely to be
    **unnecessarily** risky, i.e., whether it is highly likely that a different
    combination of hyper-parameters would have led to model with similar or
    better accuracy on the task but with lower membership inference risk.

    The rules applied from an experimental study using a grid search in which:
    - max_features was one-hot encoded from the set [None, log2, sqrt]
    - splitter was encoded using 0=best, 1=random

    The target models created were then subject to membership inference attacks
    (MIA) and the hyper-param combinations rank-ordered according to MIA AUC.
    Then a decision tree trained to recognise whether hyper-params combintions
    were in the 20% most risky.  The rules below were extracted from that tree
    for the 'least risky' nodes.

    Parameters
    ----------
    model : BaseEstimator
        Model to check for risk.

    Returns
    -------
    bool
        True if high risk, otherwise False.
    """
    unnecessary_risk: bool = False
    if isinstance(model, DecisionTreeClassifier):
        unnecessary_risk = _get_unnecessary_risk_dt(model)
    elif isinstance(model, RandomForestClassifier):
        unnecessary_risk = _get_unnecessary_risk_rf(model)
    elif isinstance(model, XGBClassifier):
        unnecessary_risk = _get_unnecessary_risk_xgb(model)
    return unnecessary_risk


def _get_unnecessary_risk_dt(model: DecisionTreeClassifier) -> bool:
    """Return whether DecisionTreeClassifier parameters are high risk."""
    max_depth = float(model.max_depth) if model.max_depth else 500
    max_features = model.max_features
    min_samples_leaf = model.min_samples_leaf
    min_samples_split = model.min_samples_split
    splitter = model.splitter
    return (
        (max_depth > 7.5 and min_samples_leaf <= 7.5 and min_samples_split <= 15)
        or (
            splitter == "best"
            and max_depth > 7.5
            and min_samples_leaf <= 7.5
            and min_samples_split > 15
        )
        or (
            splitter == "best"
            and max_depth > 7.5
            and 7.5 < min_samples_leaf <= 15
            and max_features is None
        )
        or (
            splitter == "best"
            and 3.5 < max_depth <= 7.5
            and max_features is None
            and min_samples_leaf <= 7.5
        )
        or (
            splitter == "random"
            and max_depth > 7.5
            and min_samples_leaf <= 7.5
            and max_features is None
        )
    )


def _get_unnecessary_risk_rf(model: RandomForestClassifier) -> bool:
    """Return whether RandomForestClassifier parameters are high risk."""
    max_depth = float(model.max_depth) if model.max_depth else 500
    n_estimators = model.n_estimators
    max_features = model.max_features
    min_samples_leaf = model.min_samples_leaf
    min_samples_split = model.min_samples_split
    return (
        (max_depth > 3.5 and n_estimators > 35 and max_features is not None)
        or (
            max_depth > 3.5
            and n_estimators > 35
            and min_samples_split <= 15
            and max_features is None
            and model.bootstrap
        )
        or (
            max_depth > 7.5
            and 15 < n_estimators <= 35
            and min_samples_leaf <= 15
            and not model.bootstrap
        )
    )


def _get_unnecessary_risk_xgb(model: XGBClassifier) -> bool:
    """Return whether XGBClassifier parameters are high risk.

    Check whether params exist and using xgboost defaults if not using defaults
    from https://github.com/dmlc/xgboost/blob/master/python-package/xgboost/sklearn.py
    and here: https://xgboost.readthedocs.io/en/stable/parameter.html
    """
    n_estimators = int(model.n_estimators) if model.n_estimators else 100
    max_depth = float(model.max_depth) if model.max_depth else 6
    min_child_weight = float(model.min_child_weight) if model.min_child_weight else 1.0
    return (
        (max_depth > 3.5 and 3.5 < n_estimators <= 12.5 and min_child_weight <= 1.5)
        or (max_depth > 3.5 and n_estimators > 12.5 and min_child_weight <= 3)
        or (max_depth > 3.5 and n_estimators > 62.5 and 3 < min_child_weight <= 6)
    )


def get_tree_parameter_count(dtree: DecisionTreeClassifier) -> int:
    """Read the tree structure and return the number of learned parameters."""
    n_nodes = dtree.tree_.node_count
    left = dtree.tree_.children_left
    right = dtree.tree_.children_right
    is_leaf = np.zeros(n_nodes, dtype=int)
    for node_id in range(n_nodes):
        if left[node_id] == right[node_id]:
            is_leaf[node_id] = 1
    n_leaves = is_leaf.sum()

    # degrees of freedom
    n_internal_nodes = n_nodes - n_leaves
    n_params = 2 * n_internal_nodes  # feature id and threshold
    n_params += n_leaves * (dtree.n_classes_ - 1)  # probability distribution
    return n_params


def get_model_param_count(model: BaseEstimator) -> int:
    """Return the number of trained parameters in a model."""
    n_params: int = 0
    if isinstance(model, DecisionTreeClassifier):
        n_params = _get_model_param_count_dt(model)
    elif isinstance(model, RandomForestClassifier):
        n_params = _get_model_param_count_rf(model)
    elif isinstance(model, AdaBoostClassifier):
        n_params = _get_model_param_count_ada(model)
    elif isinstance(model, XGBClassifier):
        n_params = _get_model_param_count_xgb(model)
    elif isinstance(model, MLPClassifier):
        n_params = _get_model_param_count_mlp(model)
    return n_params


def _get_model_param_count_dt(model: DecisionTreeClassifier) -> int:
    """Return the number of trained DecisionTreeClassifier parameters."""
    return get_tree_parameter_count(model)


def _get_model_param_count_rf(model: RandomForestClassifier) -> int:
    """Return the number of trained RandomForestClassifier parameters."""
    n_params: int = 0
    for member in model.estimators_:
        n_params += get_tree_parameter_count(member)
    return n_params


def _get_model_param_count_ada(model: AdaBoostClassifier) -> int:
    """Return the number of trained AdaBoostClassifier parameters."""
    n_params: int = 0
    try:  # sklearn v1.2+
        base = model.estimator
    except AttributeError:  # sklearn version <1.2
        base = model.base_estimator
    if isinstance(base, DecisionTreeClassifier):
        for member in model.estimators_:
            n_params += get_tree_parameter_count(member)
    return n_params


def _get_model_param_count_xgb(model: XGBClassifier) -> int:
    """Return the number of trained XGBClassifier parameters."""
    df = model.get_booster().trees_to_dataframe()
    n_trees = df["Tree"].max()
    total = len(df)
    n_leaves = len(df[df.Feature == "Leaf"])
    # 2 per internal node, one per clas in leaves, one weight per tree
    return 2 * (total - n_leaves) + (model.n_classes_ - 1) * n_leaves + n_trees


def _get_model_param_count_mlp(model: MLPClassifier) -> int:
    """Return the number of trained MLPClassifier parameters."""
    weights = model.coefs_  # dtype is list of numpy.ndarrays
    biasses = model.intercepts_  # dtype is list of numpy.ndarrays
    return sum(a.size for a in weights) + sum(a.size for a in biasses)


class StructuralAttack(Attack):
    """Structural attacks based on the static structure of a model."""

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        output_dir: str = "outputs",
        write_report: bool = True,
        risk_appetite_config: str = "default",
    ) -> None:
        """Construct an object to execute a structural attack.

        Parameters
        ----------
        output_dir : str
            Name of a directory to write outputs.
        write_report : bool
            Whether to generate a JSON and PDF report.
        risk_appetite_config : str
            Path to yaml file specifying TRE risk appetite.
        """
        super().__init__(output_dir=output_dir, write_report=write_report)
        self.target: Target = None
        # disclosure risk
        self.k_anonymity_risk: bool = False
        self.dof_risk: bool = False
        self.unnecessary_risk: bool = False
        self.class_disclosure_risk: bool = False
        self.lowvals_cd_risk: bool = False
        # make dummy acro object and use it to extract risk appetite
        myacro = ACRO(risk_appetite_config)
        self.risk_appetite_config = risk_appetite_config
        self.THRESHOLD = myacro.config["safe_threshold"]
        self.DOF_THRESHOLD = myacro.config["safe_dof_threshold"]
        logger.info(
            "Thresholds for count %i and Dof %i", self.THRESHOLD, self.DOF_THRESHOLD
        )
        del myacro

        # metrics
        self.attack_metrics = [
            "dof_risk",
            "k_anonymity_risk",
            "class_disclosure_risk",
            "lowvals_cd_risk",
            "unnecessary_risk",
        ]
        self.yprobs = []

    def __str__(self) -> str:
        """Return the name of the attack."""
        return "Structural attack"

    def attack(self, target: Target) -> dict:
        """Run structural attack.

        To be used when code has access to Target class and trained target model.

        Parameters
        ----------
        target : attacks.target.Target
            target as a Target class object

        Returns
        -------
        dict
            Attack report.
        """
        self.target = target
        if target.model is None:
            errstr = (
                "cannot currently call StructuralAttack.attack() "
                "unless the target contains a trained model"
            )
            raise NotImplementedError(errstr)

        # get proba values for training data
        x = self.target.X_train
        y = self.target.y_train
        assert x.shape[0] == len(y), "length mismatch between trainx and trainy"
        self.yprobs = self.target.model.predict_proba(x)

        # only equivalence classes and membership once as potentially slow
        if isinstance(target.model, DecisionTreeClassifier):
            equiv = self.dt_get_equivalence_classes()
        else:
            equiv = self.get_equivalence_classes()
        equiv_classes = equiv[0]
        equiv_counts = equiv[1]
        equiv_members = equiv[2]
        errstr = "len mismatch between equiv classes and "
        assert len(equiv_classes) == len(equiv_counts), errstr + "counts"
        assert len(equiv_classes) == len(equiv_members), errstr + "membership"

        # now assess the risk
        # Degrees of Freedom
        n_params = get_model_param_count(target.model)
        residual_dof = self.target.X_train.shape[0] - n_params
        self.dof_risk = residual_dof < self.DOF_THRESHOLD

        # k-anonymity
        mink = np.min(np.array(equiv_counts))
        self.k_anonymity_risk = mink < self.THRESHOLD

        # unnecessary risk arising from poor hyper-parameter combination.
        self.unnecessary_risk = get_unnecessary_risk(self.target.model)

        # class disclosure
        freqs = np.zeros(equiv_classes.shape)
        for group in range(freqs.shape[0]):
            freqs = equiv_classes[group] * equiv_counts[group]
        self.class_disclosure_risk = np.any(freqs < self.THRESHOLD)
        freqs[freqs == 0] = 100
        self.lowvals_cd_risk = np.any(freqs < self.THRESHOLD)

        # create the report
        output = self._make_report(target)
        # write the report
        self._write_report(output)
        # return the report
        return output

    def dt_get_equivalence_classes(self) -> tuple:
        """Get details of equivalence classes based on white box inspection."""
        destinations = self.target.model.apply(self.target.X_train)
        ret_tuple = np.unique(destinations, return_counts=True)
        leaves = ret_tuple[0]
        counts = ret_tuple[1]
        members = []
        for leaf in leaves:
            ingroup = np.asarray(destinations == leaf).nonzero()[0]
            members.append(ingroup)

        equiv_classes = np.zeros((len(leaves), self.target.model.n_classes_))
        for group in range(len(leaves)):
            sample_id = members[group][0]
            sample = self.target.X_train[sample_id]
            proba = self.target.model.predict_proba(sample.reshape(1, -1))
            equiv_classes[group] = proba
        return [equiv_classes, counts, members]

    def get_equivalence_classes(self) -> tuple:
        """Get details of equivalence classes based on predicted probabilities."""
        uniques = np.unique(self.yprobs, axis=0, return_counts=True)
        equiv_classes = uniques[0]
        equiv_counts = uniques[1]
        members = []
        for prob_vals in equiv_classes:
            ingroup = np.unique(np.asarray(self.yprobs == prob_vals).nonzero()[0])
            members.append(ingroup)
        return [equiv_classes, equiv_counts, members]

    def _get_global_metrics(self, attack_metrics: list) -> dict:
        """Get dictionary summarising metrics from a metric list.

        Parameters
        ----------
        attack_metrics : list
            List of attack metrics to be reported.

        Returns
        -------
        global_metrics : dict
            Dictionary of summary metrics.
        """
        global_metrics = {}
        if attack_metrics is not None and len(attack_metrics) != 0:
            global_metrics["dof_risk"] = self.dof_risk
            global_metrics["k_anonymity_risk"] = self.k_anonymity_risk
            global_metrics["class_disclosure_risk"] = self.class_disclosure_risk
            global_metrics["unnecessary_risk"] = self.unnecessary_risk
            global_metrics["lowvals_cd_risk"] = self.lowvals_cd_risk
        return global_metrics

    def _construct_metadata(self):
        """Construct the metadata object, after attacks."""
        super()._construct_metadata()
        self.metadata["global_metrics"] = self._get_global_metrics(self.attack_metrics)

    def _get_attack_metrics_instances(self) -> dict:
        """Construct the metadata object, after attacks."""
        attack_metrics_experiment = {}
        attack_metrics_instances = {}
        attack_metrics_experiment["attack_instance_logger"] = attack_metrics_instances
        attack_metrics_experiment["dof_risk"] = self.dof_risk
        attack_metrics_experiment["k_anonymity_risk"] = self.k_anonymity_risk
        attack_metrics_experiment["class_disclosure_risk"] = self.class_disclosure_risk
        attack_metrics_experiment["unnecessary_risk"] = self.unnecessary_risk
        attack_metrics_experiment["lowvals_cd_risk"] = self.lowvals_cd_risk
        return attack_metrics_experiment

    def _make_pdf(self, output: dict) -> None:
        attack_name: str = output["metadata"]["attack_name"]
        logger.info("PDF report not yet implemented for %s", attack_name)



"""Run a worst case attack based upon predictive probabilities."""

from __future__ import annotations

import logging
from collections.abc import Iterable

import numpy as np
from fpdf import FPDF
from sklearn.metrics import confusion_matrix
from sklearn.model_selection import train_test_split

from sacroml import metrics
from sacroml.attacks import report
from sacroml.attacks.attack import Attack, get_class_by_name
from sacroml.attacks.target import Target

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

P_THRESH = 0.05


class WorstCaseAttack(Attack):  # pylint: disable=too-many-instance-attributes
    """Worst case attack."""

    def __init__(  # pylint: disable = too-many-arguments
        self,
        output_dir: str = "outputs",
        write_report: bool = True,
        n_reps: int = 10,
        reproduce_split: int | Iterable[int] | None = 5,
        p_thresh: float = 0.05,
        n_dummy_reps: int = 1,
        train_beta: int = 1,
        test_beta: int = 1,
        test_prop: float = 0.2,
        include_model_correct_feature: bool = False,
        sort_probs: bool = True,
        attack_model: str = "sklearn.ensemble.RandomForestClassifier",
        attack_model_params: dict | None = None,
    ) -> None:
        """Construct an object to execute a worst case attack.

        Parameters
        ----------
        output_dir : str
            Name of the directory where outputs are stored.
        write_report : bool
            Whether to generate a JSON and PDF report.
        n_reps : int
            Number of attacks to run -- in each iteration an attack model
            is trained on a different subset of the data.
        reproduce_split : int or Iterable[int] or None
            Variable that controls the reproducibility of the data split.
            It can be an integer or a list of integers of length `n_reps`.
            Default : 5.
        p_thresh : float
            Threshold to determine significance of things. For instance
            `auc_p_value` and `pdif_vals`.
        n_dummy_reps : int
            Number of baseline (dummy) experiments to do.
        train_beta : int
            Value of b for beta distribution used to sample the in-sample
            (training) probabilities.
        test_beta : int
            Value of b for beta distribution used to sample the out-of-sample
            (test) probabilities.
        test_prop : float
            Proportion of data to use as a test set for the attack model.
        include_model_correct_feature : bool
            Inclusion of additional feature to hold whether or not the target model
            made a correct prediction for each example.
        sort_probs : bool
            Whether to sort combined preds (from training and test)
            to have highest probabilities in the first column.
        attack_model : str
            Class name of the attack model.
        attack_model_params : dict or None
            Dictionary of hyperparameters for the `attack_model`
            such as `min_sample_split`, `min_samples_leaf`, etc.
        """
        super().__init__(output_dir=output_dir, write_report=write_report)
        self.n_reps: int = n_reps
        self.reproduce_split: int | Iterable[int] | None = reproduce_split
        self.p_thresh: float = p_thresh
        self.n_dummy_reps: int = n_dummy_reps
        self.train_beta: int = train_beta
        self.test_beta: int = test_beta
        self.test_prop: float = test_prop
        self.include_model_correct_feature: bool = include_model_correct_feature
        self.sort_probs: bool = sort_probs
        self.attack_model: str = attack_model
        self.attack_model_params: dict | None = attack_model_params
        self.dummy_attack_metrics: list = []

    def __str__(self) -> str:
        """Return name of attack."""
        return "WorstCase attack"

    def attack(self, target: Target) -> dict:
        """Run worst case attack.

        Parameters
        ----------
        target : attacks.target.Target
            target as a Target class object

        Returns
        -------
        dict
            Attack report.
        """
        train_c = None
        test_c = None
        # compute target model probas if possible
        if (
            target.model is not None
            and target.X_train is not None
            and target.y_train is not None
        ):
            proba_train = target.model.predict_proba(target.X_train)
            proba_test = target.model.predict_proba(target.X_test)
            if self.include_model_correct_feature:
                train_c = 1 * (target.y_train == target.model.predict(target.X_train))
                test_c = 1 * (target.y_test == target.model.predict(target.X_test))
        # use supplied target model probas if unable to compute
        elif target.proba_train is not None and target.proba_test is not None:
            proba_train = target.proba_train
            proba_test = target.proba_test
        # cannot proceed
        else:
            logger.info("Insufficient Target details to run worst case attack.")
            return {}
        # execute attack
        self.attack_from_preds(
            proba_train,
            proba_test,
            train_correct=train_c,
            test_correct=test_c,
        )
        # create the report
        output = self._make_report(target)
        # write the report
        self._write_report(output)
        # return the report
        return output

    def _make_report(self, target: Target) -> dict:
        """Create attack report."""
        output = super()._make_report(target)
        output["dummy_attack_experiments_logger"] = (
            self._get_dummy_attack_metrics_experiments_instances()
        )
        return output

    def attack_from_preds(
        self,
        proba_train: np.ndarray,
        proba_test: np.ndarray,
        train_correct: np.ndarray | None = None,
        test_correct: np.ndarray | None = None,
    ) -> None:
        """Run attack based upon the predictions in proba_train and proba_test.

        Parameters
        ----------
        proba_train : np.ndarray
            Array of train predictions. One row per example, one column per class.
        proba_test : np.ndarray
            Array of test predictions. One row per example, one column per class.
        """
        logger.info("Running main attack repetitions")
        attack_metric_dict = self.run_attack_reps(
            proba_train,
            proba_test,
            train_correct=train_correct,
            test_correct=test_correct,
        )
        self.attack_metrics = attack_metric_dict["mia_metrics"]

        self.dummy_attack_metrics = []
        if self.n_dummy_reps > 0:
            logger.info("Running dummy attack reps")
            n_train_rows = len(proba_train)
            n_test_rows = len(proba_test)
            for _ in range(self.n_dummy_reps):
                d_train_preds, d_test_preds = self.generate_arrays(
                    n_train_rows,
                    n_test_rows,
                    self.train_beta,
                    self.test_beta,
                )
                temp_attack_metric_dict = self.run_attack_reps(
                    d_train_preds, d_test_preds
                )
                temp_metrics = temp_attack_metric_dict["mia_metrics"]
                self.dummy_attack_metrics.append(temp_metrics)

        logger.info("Finished running attacks")

    def _prepare_attack_data(
        self,
        proba_train: np.ndarray,
        proba_test: np.ndarray,
        train_correct: np.ndarray = None,
        test_correct: np.ndarray = None,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Prepare training data and labels for attack model.

        Combines the train and test preds into a single numpy array
        (optionally) sorting each row to have the highest probabilities in the
        first column. Constructs a label array that has ones corresponding to
        training rows and zeros to testing rows.
        """
        if self.sort_probs:
            logger.info("Sorting probabilities to leave highest value in first column")
            proba_train = -np.sort(-proba_train, axis=1)
            proba_test = -np.sort(-proba_test, axis=1)

        logger.info("Creating MIA data")

        if self.include_model_correct_feature and train_correct is not None:
            proba_train = np.hstack((proba_train, train_correct[:, None]))
            proba_test = np.hstack((proba_test, test_correct[:, None]))

        mi_x = np.vstack((proba_train, proba_test))
        mi_y = np.hstack((np.ones(len(proba_train)), np.zeros(len(proba_test))))
        return (mi_x, mi_y)

    def _get_attack_model(self):
        """Return an instantiated attack model."""
        # load attack model module and get class
        model = get_class_by_name(self.attack_model)
        params = self.attack_model_params
        if (  # set custom default parameters for RF attack model
            self.attack_model == "sklearn.ensemble.RandomForestClassifier"
            and self.attack_model_params is None
        ):
            params = {
                "min_samples_split": 20,
                "min_samples_leaf": 10,
                "max_depth": 5,
            }
        # instantiate attack model
        return model(**params) if params is not None else model()

    def _get_reproducible_split(self) -> list:
        """Return a list of splits."""
        split = self.reproduce_split
        n_reps = self.n_reps
        if isinstance(split, int):
            split = [split] + [x**2 for x in range(split, split + n_reps - 1)]
        else:
            # remove potential duplicates
            split = list(dict.fromkeys(split))
            if len(split) == n_reps:
                pass
            elif len(split) > n_reps:
                print("split", split, "nreps", n_reps)
                split = list(split)[0:n_reps]
                print(
                    "WARNING: the length of the parameter 'reproduce_split' "
                    "is longer than n_reps. Values have been removed."
                )
            else:
                # assign values to match length of n_reps
                split += [split[-1] * x for x in range(2, (n_reps - len(split) + 2))]
                print(
                    "WARNING: the length of the parameter 'reproduce_split' "
                    "is shorter than n_reps. Values have been added."
                )
            print("reproduce split now", split)
        return split

    def run_attack_reps(  # pylint: disable = too-many-locals
        self,
        proba_train: np.ndarray,
        proba_test: np.ndarray,
        train_correct: np.ndarray = None,
        test_correct: np.ndarray = None,
    ) -> dict:
        """Run actual attack reps from train and test predictions.

        Parameters
        ----------
        proba_train : np.ndarray
            Predictions from the model on training (in-sample) data.
        proba_test : np.ndarray
            Predictions from the model on testing (out-of-sample) data.

        Returns
        -------
        dict
            Dictionary of mia_metrics (a list of metric across repetitions).
        """
        mi_x, mi_y = self._prepare_attack_data(
            proba_train, proba_test, train_correct, test_correct
        )

        mia_metrics = []
        split = self._get_reproducible_split()

        for rep in range(self.n_reps):
            logger.info("Rep %d of %d split %d", rep + 1, self.n_reps, split[rep])

            mi_train_x, mi_test_x, mi_train_y, mi_test_y = train_test_split(
                mi_x,
                mi_y,
                test_size=self.test_prop,
                stratify=mi_y,
                random_state=split[rep],
                shuffle=True,
            )

            attack_classifier = self._get_attack_model()
            attack_classifier.fit(mi_train_x, mi_train_y)

            y_pred_proba = attack_classifier.predict_proba(mi_test_x)
            mia_metrics.append(metrics.get_metrics(y_pred_proba, mi_test_y))

            if self.include_model_correct_feature and train_correct is not None:
                # Compute the Yeom TPR and FPR
                yeom_preds = mi_test_x[:, -1]
                tn, fp, fn, tp = confusion_matrix(mi_test_y, yeom_preds).ravel()
                mia_metrics[-1]["yeom_tpr"] = tp / (tp + fn)
                mia_metrics[-1]["yeom_fpr"] = fp / (fp + tn)
                mia_metrics[-1]["yeom_advantage"] = (
                    mia_metrics[-1]["yeom_tpr"] - mia_metrics[-1]["yeom_fpr"]
                )

        logger.info("Finished simulating attacks")
        return {"mia_metrics": mia_metrics}

    def _get_global_metrics(self, attack_metrics: list) -> dict:
        """Summarise metrics from a metric list.

        Parameters
        ----------
        attack_metrics : List
            list of attack metrics dictionaries

        Returns
        -------
        global_metrics : Dict
            Dictionary of summary metrics
        """
        global_metrics = {}
        if attack_metrics is not None and len(attack_metrics) != 0:
            auc_p_vals = [
                metrics.auc_p_val(
                    m["AUC"], m["n_pos_test_examples"], m["n_neg_test_examples"]
                )[0]
                for m in attack_metrics
            ]

            m = attack_metrics[0]
            _, auc_std = metrics.auc_p_val(
                0.5, m["n_pos_test_examples"], m["n_neg_test_examples"]
            )

            global_metrics["null_auc_3sd_range"] = (
                f"{0.5 - 3*auc_std:.4f} -> {0.5 + 3*auc_std:.4f}"
            )
            global_metrics["n_sig_auc_p_vals"] = self._get_n_significant(
                auc_p_vals, self.p_thresh
            )
            global_metrics["n_sig_auc_p_vals_corrected"] = self._get_n_significant(
                auc_p_vals, self.p_thresh, bh_fdr_correction=True
            )

            pdif_vals = [np.exp(-m["PDIF01"]) for m in attack_metrics]
            global_metrics["n_sig_pdif_vals"] = self._get_n_significant(
                pdif_vals, self.p_thresh
            )
            global_metrics["n_sig_pdif_vals_corrected"] = self._get_n_significant(
                pdif_vals, self.p_thresh, bh_fdr_correction=True
            )

        return global_metrics

    def _get_n_significant(
        self, p_val_list: list[float], p_thresh: float, bh_fdr_correction: bool = False
    ) -> int:
        """Return number of p-values significant at `p_thresh`.

        Can perform multiple testing correction.
        """
        if not bh_fdr_correction:
            return sum(1 for p in p_val_list if p <= p_thresh)
        p_val_list = np.asarray(sorted(p_val_list))
        n_vals = len(p_val_list)
        hoch_vals = np.array([(k / n_vals) * P_THRESH for k in range(1, n_vals + 1)])
        bh_sig_list = p_val_list <= hoch_vals
        return np.where(bh_sig_list)[0].max() + 1 if any(bh_sig_list) else 0

    def _generate_array(self, n_rows: int, beta: float) -> np.ndarray:
        """Generate array of predictions, used when doing baseline experiments.

        Parameters
        ----------
        n_rows : int
            The number of rows worth of data to generate.
        beta : float
            The beta parameter for sampling probabilities.

        Returns
        -------
        preds : np.ndarray
            Array of predictions. Two columns, `n_rows` rows.
        """
        preds = np.zeros((n_rows, 2), float)
        for row_idx in range(n_rows):
            train_class = np.random.choice(2)
            train_prob = np.random.beta(1, beta)
            preds[row_idx, train_class] = train_prob
            preds[row_idx, 1 - train_class] = 1 - train_prob
        return preds

    def generate_arrays(
        self,
        n_rows_in: int,
        n_rows_out: int,
        train_beta: float = 2,
        test_beta: float = 2,
    ) -> tuple[np.ndarray, np.ndarray]:
        """Generate train and test prediction arrays, used when computing baseline.

        Parameters
        ----------
        n_rows_in : int
            Number of rows of in-sample (training) probabilities.
        n_rows_out : int
            Number of rows of out-of-sample (testing) probabilities.
        train_beta : float
            Beta value for generating train probabilities.
        test_beta : float:
            Beta value for generating test probabilities.

        Returns
        -------
        proba_train : np.ndarray
            Array of train predictions (n_rows x 2 columns).
        proba_test : np.ndarray
            Array of test predictions (n_rows x 2 columns).
        """
        proba_train = self._generate_array(n_rows_in, train_beta)
        proba_test = self._generate_array(n_rows_out, test_beta)
        return proba_train, proba_test

    def _construct_metadata(self) -> None:
        """Construct the metadata object after attacks."""
        super()._construct_metadata()

        self.metadata["global_metrics"] = self._get_global_metrics(self.attack_metrics)
        self.metadata["baseline_global_metrics"] = self._get_global_metrics(
            self._unpack_dummy_attack_metrics_experiments_instances()
        )

    def _unpack_dummy_attack_metrics_experiments_instances(self) -> list:
        """Construct the metadata object after attacks."""
        dummy_attack_metrics_instances = []
        for exp_rep, _ in enumerate(self.dummy_attack_metrics):
            temp_dummy_attack_metrics = self.dummy_attack_metrics[exp_rep]
            dummy_attack_metrics_instances += temp_dummy_attack_metrics
        return dummy_attack_metrics_instances

    def _get_attack_metrics_instances(self) -> dict:
        """Construct the metadata object after attacks."""
        attack_metrics_experiment = {}
        attack_metrics_instances = {}
        for rep, _ in enumerate(self.attack_metrics):
            attack_metrics_instances["instance_" + str(rep)] = self.attack_metrics[rep]
        attack_metrics_experiment["attack_instance_logger"] = attack_metrics_instances
        return attack_metrics_experiment

    def _get_dummy_attack_metrics_experiments_instances(self) -> dict:
        """Construct the metadata object after attacks."""
        dummy_attack_metrics_experiments = {}
        for exp_rep, _ in enumerate(self.dummy_attack_metrics):
            temp_dummy_attack_metrics = self.dummy_attack_metrics[exp_rep]
            dummy_attack_metric_instances = {}
            for rep, _ in enumerate(temp_dummy_attack_metrics):
                dummy_attack_metric_instances["instance_" + str(rep)] = (
                    temp_dummy_attack_metrics[rep]
                )
            temp = {}
            temp["attack_instance_logger"] = dummy_attack_metric_instances
            dummy_attack_metrics_experiments[
                "dummy_attack_metrics_experiment_" + str(exp_rep)
            ] = temp
        return dummy_attack_metrics_experiments

    def _make_pdf(self, output: dict) -> FPDF:
        """Create PDF report."""
        return report.create_mia_report(output)



"""Code for automatic report generation."""

import json
import os
from typing import Any

import numpy as np
import pylab as plt
from fpdf import FPDF
from pypdf import PdfWriter

from sacroml.attacks.attack_report_formatter import GenerateJSONModule

# Adds a border to all pdf cells of set to 1 -- useful for debugging
BORDER = 0

DISPLAY_METRICS = [
    "AUC",
    "ACC",
    "Advantage",
    "FDIF01",
    "PDIF01",
    "TPR@0.1",
    "TPR@0.01",
    "TPR@0.001",
    "TPR@1e-05",
]

MAPPINGS = {"PDIF01": lambda x: np.exp(-x)}

INTRODUCTION = (
    "This report provides a summary of a series of simulated attack experiments "
    "performed on the model outputs provided. An attack model is trained to "
    "attempt to distinguish between outputs from training (in-sample) and "
    "testing (out-of-sample) data. The metrics below describe the success of "
    "this classifier. A successful classifier indicates that the original model "
    "is unsafe and should not be allowed to be released from the TRE.\n In "
    "particular, the simulation splits the data provided into test and train "
    "sets (each will in- and out-of-sample examples). The classifier is trained "
    "on the train set and evaluated on the test set. This is repeated with "
    "different train/test splits a user-specified number of times.\n To help "
    "place the results in context, the code may also have run a series of "
    "baseline experiments. In these, random model outputs for hypothetical in- "
    "and out-of-sample data are generated with identical statistical properties. "
    "In these baseline cases, there is no signal that an attacker could leverage "
    "and therefore these values provide a baseline against which the actual "
    "values can be compared.\n For some metrics (FDIF and AUC), we are able to "
    "compute p-values. In each case, shown below (in the Global metrics "
    "sections) is the number of repetitions that exceeded the p-value threshold "
    "both without, and with correction for multiple testing (Benjamini-Hochberg "
    "procedure).\n ROC curves for all real (red) and dummy (blue) repetitions are "
    "provided. These are shown in log space (as reommended here [ADD URL]) to "
    "emphasise the region in which risk is highest -- the bottom left (are high "
    "true positive rates possible with low false positive rates).\n A description "
    "of the metrics and how to interpret them within the context of an attack is "
    "given below."
)

LOGROC_CAPTION = (
    "This plot shows the False Positive Rate (x) versus the True Positive Rate "
    "(y). The axes are in log space enabling us to focus on areas where the "
    "False Positive Rate is low (left hand area). Curves above the y = x line "
    "(black dashes) in this region represent a disclosure risk as an attacker "
    "can obtain many more true than false positives. The solid coloured lines "
    "show the curves for the attack simulations with the true model outputs. The "
    "lighter grey lines show the curves for randomly generated outputs with no "
    "structure (i.e. in- and out-of- sample predictions are generated from the "
    "same distributions. Solid curves consistently higher than the "
    "grey curves in the left hand part of the plot are a sign of concern. "
)

GLOSSARY = {
    "AUC": "Area Under the ROC curve",
    "True Positive Rate (TPR)": (
        "The true positive rate is the number of True Positives that are "
        "predicted as positive as a proportion of the total number of positives. "
        "If an attacker has N examples that were actually in the training set, "
        "the TPR is the proportion of these that they predict as being in the "
        "training set."
    ),
    "ACC": "The proportion of predictions that the attacker makes that are correct.",
}


def write_json(output: dict, dest: str) -> None:
    """Write attack report to JSON."""
    attack_formatter = GenerateJSONModule(dest + ".json")
    attack_report: str = json.dumps(output, cls=CustomJSONEncoder)
    attack_name: str = output["metadata"]["attack_name"]
    attack_formatter.add_attack_output(attack_report, attack_name)


class CustomJSONEncoder(json.JSONEncoder):
    """JSON encoder that can cope with numpy arrays, etc."""

    def default(self, o: Any):
        """If an object is an np.ndarray, convert to list."""
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, (np.int64, np.int32)):
            return int(o)
        if isinstance(o, np.bool_):
            return bool(o)
        try:  # Try the default method first
            return super().default(o)
        except TypeError:
            return str(o)  # If object is not serializable, convert it to a string


def _write_dict(pdf: FPDF, data: dict, border: int = BORDER) -> None:
    """Write a dictionary to the pdf."""
    for key, value in data.items():
        pdf.set_font("arial", "B", 14)
        pdf.cell(0, 5, key, border, 1, "L")
        pdf.set_font("arial", "", 12)
        pdf.multi_cell(0, 5, str(value), 0, 1)
        pdf.ln(h=5)


def title(
    pdf: FPDF,
    text: str,
    border: int = BORDER,
    font_size: int = 24,
    font_style: str = "B",
) -> None:
    """Write a title block."""
    pdf.set_font("arial", font_style, font_size)
    pdf.ln(h=5)
    pdf.cell(0, 0, text, border, 1, "C")
    pdf.ln(h=5)


def subtitle(  # pylint: disable = too-many-arguments
    pdf: FPDF,
    text: str,
    indent: int = 10,
    border: int = BORDER,
    font_size: int = 12,
    font_style: str = "B",
) -> None:
    """Write a subtitle block."""
    pdf.cell(indent, border=border)
    pdf.set_font("arial", font_style, font_size)
    pdf.cell(75, 10, text, border, 1)


def line(  # pylint: disable = too-many-arguments
    pdf: FPDF,
    text: str,
    indent: int = 0,
    border: int = BORDER,
    font_size: int = 11,
    font_style: str = "",
    font: str = "arial",
) -> None:
    """Write a standard block."""
    if indent > 0:
        pdf.cell(indent, border=border)
    pdf.set_font(font, font_style, font_size)
    pdf.multi_cell(0, 5, text, border, 1)


def _roc_plot_single(metrics: dict, save_name: str) -> None:
    """Create a roc_plot for a single experiment."""
    plt.figure()
    plt.plot([0, 1], [0, 1], "k--")
    plt.plot(metrics["fpr"], metrics["tpr"], "r", linewidth=2)
    plt.xscale("log")
    plt.yscale("log")
    plt.grid()
    plt.ylabel("True Positive Rate")
    plt.xlabel("False Positive Rate")
    plt.tight_layout()
    plt.savefig(save_name)


def _roc_plot(metrics: dict, save_name: str) -> None:
    """Create a roc plot for multiple repetitions."""
    plt.figure()
    plt.plot([0, 1], [0, 1], "k--")

    # Compute average ROC
    base_fpr = np.linspace(0, 1, 1000)
    all_tpr = np.zeros((len(metrics), len(base_fpr)), float)
    for i, metric_set in enumerate(metrics):
        all_tpr[i, :] = np.interp(base_fpr, metric_set["fpr"], metric_set["tpr"])

    for _, metric_set in enumerate(metrics):
        plt.plot(
            metric_set["fpr"], metric_set["tpr"], color="lightsalmon", linewidth=0.5
        )

    tpr_mu = all_tpr.mean(axis=0)
    plt.plot(base_fpr, tpr_mu, "r")
    plt.xscale("log")
    plt.yscale("log")
    plt.xlabel("False Positive Rate")
    plt.ylabel("True Positive Rate")
    plt.tight_layout()
    plt.grid()
    plt.savefig(save_name)


def create_mia_report(attack_output: dict) -> FPDF:
    """Make a worst case membership inference report.

    Parameters
    ----------
    attack_output : dict
        Dictionary with the following items:

        metadata : dict
            Dictionary of metadata.
        attack_experiment_logger : dict
            List of metrics as dictionary items for an experiment.
        dummy_attack_experiment_logger : dict
            List of metrics as dictionary items across dummy experiments.

    Returns
    -------
    pdf : fpdf.FPDF
        fpdf document object
    """
    mia_metrics = [
        v
        for _, v in attack_output["attack_experiment_logger"][
            "attack_instance_logger"
        ].items()
    ]
    metadata = attack_output["metadata"]

    path: str = metadata["attack_params"]["output_dir"]
    dest_log_roc = os.path.join(path, "log_roc.png")
    _roc_plot(mia_metrics, dest_log_roc)

    pdf = FPDF()
    pdf.add_page()
    pdf.set_xy(0, 0)
    title(pdf, "WorstCase MIA attack result report")
    subtitle(pdf, "Introduction")
    line(pdf, INTRODUCTION)
    subtitle(pdf, "Experiment summary")
    for key, value in metadata["attack_params"].items():
        line(pdf, f"{key:>30s}: {str(value):30s}", font="courier")
    subtitle(pdf, "Global metrics")
    for key, value in metadata["global_metrics"].items():
        line(pdf, f"{key:>30s}: {str(value):30s}", font="courier")

    subtitle(pdf, "Metrics")
    line(
        pdf,
        "The following show summaries of the attack metrics over the repetitions",
        font="arial",
    )
    for metric in DISPLAY_METRICS:
        vals = np.array([m[metric] for m in mia_metrics])
        if metric in MAPPINGS:
            vals = np.array([MAPPINGS[metric](v) for v in vals])
        text = (
            f"{metric:>12} mean = {vals.mean():.2f}, var = {vals.var():.4f}, "
            f"min = {vals.min():.2f}, max = {vals.max():.2f}"
        )
        line(pdf, text, font="courier")

    _add_log_roc_to_page(dest_log_roc, pdf)
    line(pdf, LOGROC_CAPTION)

    pdf.add_page()
    title(pdf, "Glossary")
    _write_dict(pdf, GLOSSARY)

    if os.path.exists(dest_log_roc):
        os.remove(dest_log_roc)
    return pdf


def write_pdf(report_dest: str, pdf_report: FPDF) -> None:
    """Create pdf and append contents if it already exists."""
    if os.path.exists(report_dest + ".pdf"):
        old_pdf = report_dest + ".pdf"
        new_pdf = report_dest + "_new.pdf"
        pdf_report.output(new_pdf)
        merger = PdfWriter()
        for pdf in [old_pdf, new_pdf]:
            merger.append(pdf)
        merger.write(old_pdf)
        merger.close()
        os.remove(new_pdf)
    else:
        pdf_report.output(report_dest + ".pdf")


def _add_log_roc_to_page(log_roc: str = None, pdf_obj: FPDF = None) -> None:
    if log_roc is not None:
        pdf_obj.add_page()
        subtitle(pdf_obj, "Log ROC")
        pdf_obj.image(log_roc, x=None, y=None, w=0, h=140, type="", link="")
        pdf_obj.set_font("arial", "", 12)


def _plot_lira_individuals(metrics: dict, dest: str) -> None:
    """Create a plot of the individual record LiRA scores."""
    scores = np.array(metrics["individual"]["score"])
    member = np.array(metrics["individual"]["member"])

    _, axes = plt.subplots(1, 2, figsize=(12.4, 4.8))

    # members
    mask = member == 1
    y_train = scores[mask]
    x_train = np.arange(y_train.shape[0])

    sorted_indicies = np.argsort(y_train)
    y_sorted = y_train[sorted_indicies]

    axes[0].scatter(x_train, y_sorted, color="b", s=2, label="LiRA Probability")
    axes[0].set_title("Member Records")
    axes[0].set_xlabel("Record (sorted)")
    axes[0].legend(loc=0)

    # nonmembers
    y_test = scores[~mask]
    x_test = np.arange(y_test.shape[0])

    sorted_indicies = np.argsort(y_test)
    y_sorted = y_test[sorted_indicies]

    axes[1].scatter(x_test, y_sorted, color="r", s=2, label="LiRA Probability")
    axes[1].set_title("Nonmember Records")
    axes[1].set_xlabel("Record (sorted)")
    axes[1].legend(loc=0)

    plt.tight_layout()
    plt.savefig(dest)


def create_lr_report(output: dict) -> FPDF:
    """Make a lira membership inference report.

    Parameters
    ----------
    output : dict
        Dictionary with the following items:

        metadata : dict
            Dictionary of metadata.

        attack_experiment_logger : dict
            List of metrics as dictionary items for an experiments.
            In case of LiRA attack scenario, this will have dictionary items of
            `attack_instance_logger` that will have a single metrics dictionary.

    Returns
    -------
    pdf : fpdf.FPDF
        fpdf document object
    """
    mia_metrics = [
        v
        for _, v in output["attack_experiment_logger"]["attack_instance_logger"].items()
    ][0]
    metadata = output["metadata"]

    path: str = metadata["attack_params"]["output_dir"]
    dest_log_roc = os.path.join(path, "log_roc.png")
    _roc_plot_single(mia_metrics, dest_log_roc)
    pdf = FPDF()
    pdf.add_page()
    pdf.set_xy(0, 0)
    title(pdf, "Likelihood Ratio Attack Report")
    subtitle(pdf, "Introduction")
    subtitle(pdf, "Metadata")
    for key, value in metadata["attack_params"].items():
        line(pdf, f"{key:>30s}: {str(value):30s}", font="courier")
    for key, value in metadata["global_metrics"].items():
        line(pdf, f"{key:>30s}: {str(value):30s}", font="courier")
    subtitle(pdf, "Metrics")
    sub_metrics_dict = {
        key: val for key, val in mia_metrics.items() if isinstance(val, float)
    }
    for key, value in sub_metrics_dict.items():
        val = MAPPINGS[key](value) if key in MAPPINGS else value
        line(pdf, f"{key:>30s}: {val:.4f}", font="courier")

    pdf.add_page()
    subtitle(pdf, "ROC Curve")
    pdf.image(dest_log_roc, x=None, y=None, w=0, h=140, type="", link="")

    dest_ind_plot = os.path.join(path, "lira_individual.png")
    if "individual" in mia_metrics:
        _plot_lira_individuals(mia_metrics, dest_ind_plot)
        pdf.add_page()
        subtitle(pdf, "Individual LiRA Scores")
        pdf.image(dest_ind_plot, x=None, y=None, w=180, h=90, type="", link="")

    # clean up
    files = [dest_log_roc, dest_ind_plot]
    for file in files:
        if os.path.exists(file):
            os.remove(file)
    return pdf
